from copy import deepcopy
from typing import Dict, List, Literal, Union, Optional, Any
from docassemble.base.util import (
    Address,
    as_datetime,
    capitalize,
    comma_and_list,
    comma_list,
    country_name,
    DADateTime,
    DADict,
    DAFile,
    DAList,
    date_difference,
    DAWeb,
    ensure_definition,
    get_config,
    get_country,
    her,
    his,
    Individual,
    IndividualName,
    its,
    name_suffix,
    phone_number_formatted,
    phone_number_is_valid,
    showifdef,
    state_name,
    states_list,
    subdivision_type,
    their,
    this_thread,
    url_action,
    validation_error,
    word,
    your,
)
import random
import re
import pycountry

__all__ = [
    "Abuser",
    "Abuser",
    "AddressList",
    "ALAddress",
    "ALAddressList",
    "ALIndividual",
    "ALPeopleList",
    "Applicant",
    "Applicant",
    "get_visible_al_nav_items",
    "github_modified_date",
    "has_parsable_pronouns",
    "HousingAuthority",
    "is_phone_or_email",
    "Landlord",
    "language_name",
    "parse_custom_pronouns",
    "PeopleList",
    "safe_subdivision_type",
    "section_links",
    "section_links",
    "Survivor",
    "Survivor",
    "Tenant",
    "VCIndividual",
    "will_send_to_real_court",
    "is_sms_enabled",
]

##########################################################
# Base classes


def safe_subdivision_type(country_code: str) -> Optional[str]:
    """
    Returns the subdivision type for the country with the given country code.
    If no subdivision type is found, returns None.

    Args:
        country_code (str): The ISO-3166-1 alpha-2 code for the country.

    Returns:
        Optional[str]: The subdivision type for the country with the given country code.
    """
    try:
        return subdivision_type(country_code)
    except:
        return None


class ALAddress(Address):
    """
    This class is used to store addresses. The ALAddress class extends the Address
    class with the `address_fields()` method and "smarter"
    handling of the unit attribute when printing a formatted address.

    Attributes:
        address (str): The street where the person lives.
        unit (str): The unit number where the person lives.
        city (str): The city where the person lives.
        state (str): The state where the person lives.
        zip (str): The zip code where the person lives.
        country (str): The country where the person lives.
        impounded (Optional[bool]): Whether the address is impounded.
    """

    def address_fields(
        self,
        country_code: Optional[str] = None,
        default_state: Optional[str] = None,
        show_country: bool = False,
        show_county: bool = False,
        show_if: Union[str, Dict[str, str], None] = None,
        allow_no_address: bool = False,
        ask_if_impounded: Optional[bool] = False,
    ) -> List[Dict[str, Any]]:
        """
        Return a YAML structure representing the list of fields for the object's address.

        Optionally, allow the user to specify they do not have an address. When using
        `allow_no_address=True`, ensure to trigger the question with `users[0].address.has_no_address`
        rather than `users[0].address.address`. If `show_if` is used, it will not be applied when
        `allow_no_address` is also used. Ensure `country_code` adheres to ISO-3166-1 alpha-2 code standard.

        NOTE: This function is stateful under specific conditions. Refer to the conditions mentioned below.

        Args:
            country_code (Optional[str]): ISO-3166-1 alpha-2 code of the country. Defaults to None.
            default_state (Optional[str]): Default state to set. Defaults to None.
            show_country (bool): Whether to display the country field. Defaults to False.
            show_county (bool): Whether to display the county field. Defaults to False.
            show_if (Union[str, Dict[str, str], None]): Condition to display each field. Defaults to None.
            allow_no_address (bool): Allow users to specify they don't have an address. Defaults to False.
            ask_if_impounded (Optional[bool]): Whether to ask if the address is impounded. Defaults to False.

        Returns:
            list: A list of YAML structure representing address fields.

        Notes:
            - The function will set the `country` attribute of the Address to `country_code` under these
            circumstances:
                1. The `country_code` parameter is used.
                2. The `show_country` parameter is not used.
                3. `country_code` differs from the value returned by `get_country()`.

            - Link to ISO-3166-1 alpha-2 codes:
            [Officially assigned code elements](https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2#Officially_assigned_code_elements).
        """
        # make sure the state name still returns a meaningful value if the interview country
        # differs from the server's country.
        if country_code and country_code != get_country() and not show_country:
            self.country = country_code
        # Priority order for country: already answered country, passed in country_code, then get_country
        prev_selected_country = showifdef(self.attr_name("country"), None, prior=True)
        if prev_selected_country:
            country_code = prev_selected_country
        if not country_code:
            country_code = get_country()
        if allow_no_address:
            fields = [
                {
                    "label": str(self.has_no_address_label),
                    "field": self.attr_name("has_no_address"),
                    "datatype": "yesno",
                },
                {
                    "label": str(self.has_no_address_explanation_label),
                    "field": self.attr_name("has_no_address_explanation"),
                    "datatype": "area",
                    "rows": 2,
                    "help": str(self.has_no_address_explanation_help),
                    "show if": self.attr_name("has_no_address"),
                    "required": False,
                },
            ]
        else:
            fields = []
        fields.extend(
            [
                {
                    "label": str(self.address_label),
                    "address autocomplete": bool(
                        (get_config("google") or {}).get("google maps api key")
                    ),
                    "field": self.attr_name("address"),
                },
                {
                    "label": str(self.unit_label),
                    "field": self.attr_name("unit"),
                    "required": False,
                },
            ]
        )
        if allow_no_address:
            fields[-1]["hide if"] = self.attr_name("has_no_address")
            fields[-2]["hide if"] = self.attr_name("has_no_address")

        fields.append({"label": str(self.city_label), "field": self.attr_name("city")})

        if country_code and not show_country:
            fields.append(
                {
                    "label": str(self.state_label),
                    "field": self.attr_name("state"),
                    "code": "states_list(country_code='{}')".format(country_code),
                    "default": default_state if default_state else "",
                }
            )
        else:  # when you are allowed to change country
            fields.append(
                {
                    "label": str(self.state_or_province_label),
                    "field": self.attr_name("state"),
                    "default": default_state if default_state else "",
                }
            )
        if country_code == "US" and not show_country:
            fields.append(
                {
                    "label": str(self.zip_label),
                    "field": self.attr_name("zip"),
                    "required": False,
                }
            )
        else:
            fields.append(
                # We have code in ALWeaver that relies on "zip", so keep attribute same for now
                {
                    "label": str(self.postal_code_label),
                    "field": self.attr_name("zip"),
                    "required": False,
                }
            )
        if allow_no_address:
            fields[-1]["hide if"] = self.attr_name("has_no_address")

        if show_county:
            fields.append(
                {
                    "label": str(self.county_label),
                    "field": self.attr_name("county"),
                    "required": False,
                }
            )
        if show_country:
            fields.append(
                {
                    "label": str(self.country_label),
                    "field": self.attr_name("country"),
                    "required": False,
                    "code": "countries_list()",
                    "default": country_code,
                }
            )
            # NOTE: using , "datatype": "combobox" might be nice but does not play together well w/ address autocomplete
        if not allow_no_address:
            # show if isn't compatible with the hide if logic for `allow_no_address`
            if show_if:
                for field in fields:
                    field["show if"] = show_if

        if ask_if_impounded:
            fields.append(
                {
                    "label": str(self.impounded_label),
                    "field": self.attr_name("impounded"),
                    "datatype": "yesno",
                }
            )

        return fields

    def formatted_unit(
        self, language: Optional[str] = None, require: bool = False, bare: bool = False
    ) -> str:
        """
        Returns the unit, formatted appropriately.

        Args:
            language (str, optional): The language in which to format the unit. Defaults to None (which uses system language).
            require (bool, optional): A flag indicating whether the unit is required. If set to True, the function will
                                    raise an error if the unit attribute does not exist. Defaults to False.
            bare (bool, optional): A flag indicating whether to add the word 'Unit' before the unit number. If set to
                                True, the function will not add 'Unit' regardless of other conditions. Defaults to False.

        Returns:
            str: The formatted unit. If the unit attribute does not exist and require is set to False, this will be an
                empty string. If the unit attribute exists and is not None or an empty string, the function will return
                the unit number, possibly prefixed with 'Unit'. If the unit attribute exists and is None or an empty
                string, the function will return an empty string.
        """
        if (
            not hasattr(self, "unit")
            and not hasattr(self, "floor")
            and not hasattr(self, "room")
        ):
            if require:
                self.unit
            else:
                return ""
        if hasattr(self, "unit") and self.unit is not None and self.unit != "":
            unit_lower = str(self.unit).lower()
            # Sometimes people neglect to add a word before the unit number,
            # use some heuristics to decide when it's necessary to add one.
            if not bare and (
                unit_lower.isnumeric()
                or (
                    not " " in self.unit
                    and not any(
                        x in unit_lower
                        for x in [
                            "apt",
                            "unit",
                            "suite",
                            "bldg",
                            "fl",
                            "apartment",
                            "building",
                            "floor",
                            "ste",
                        ]
                    )
                )
            ):
                return word("Unit", language=language) + " " + str(self.unit)
            else:
                return str(self.unit)
        if hasattr(self, "floor") and self.floor != "" and self.floor is not None:
            return word("Floor", language=language) + " " + str(self.floor)
        if hasattr(self, "room") and self.room != "" and self.room is not None:
            return word("Room", language=language) + " " + str(self.room)
        return ""

    def block(
        self,
        language: Optional[str] = None,
        international: bool = False,
        show_country: Optional[bool] = None,
        bare: bool = False,
        long_state: bool = False,
        show_impounded: bool = False,
    ) -> str:
        """Returns a one-line formatted address, primarily for geocoding.

        Args:
            language (str, optional): Language for the address format.
            international (bool): If True, formats the address as an international address. Defaults to False.
            show_country (bool, optional): If True, includes the country in the formatted address.
                If None, decides based on the country attribute.
            bare (bool): If True, excludes certain formatting elements. Defaults to False.
            long_state (bool): If True, uses the full state name. Defaults to False.
            show_impounded (bool): If True, shows the address even if impounded. Defaults to False.

        Returns:
            str: The one-line formatted address.
        """
        if this_thread.evaluation_context == "docx":
            line_breaker = '</w:t><w:br/><w:t xml:space="preserve">'
        else:
            line_breaker = " [NEWLINE] "

        if not show_impounded and (hasattr(self, "impounded") and self.impounded):
            return str(self.impounded_output_label)
        if (
            hasattr(self, "has_no_address")
            and self.has_no_address
            and hasattr(self, "has_no_address_explanation")
        ):
            return (
                self.has_no_address_explanation
                + line_breaker
                + self.city
                + line_breaker
                + self.state
            )
        if international:
            i18n_address = {}
            if (
                (not hasattr(self, "address"))
                and hasattr(self, "street_number")
                and hasattr(self, "street")
            ):
                i18n_address["street_address"] = (
                    str(self.street_number) + " " + str(self.street)
                )
            else:
                i18n_address["street_address"] = str(self.address)
            the_unit = self.formatted_unit(language=language, bare=bare)
            if the_unit != "":
                i18n_address["street_address"] += "\n" + the_unit
            if hasattr(self, "sublocality_level_1") and self.sublocality_level_1:
                i18n_address["city_area"] = str(self.sublocality_level_1)
            i18n_address["city"] = str(self.city)
            if hasattr(self, "state") and self.state:
                i18n_address["country_area"] = str(self.state)
            if hasattr(self, "zip") and self.zip:
                i18n_address["postal_code"] = str(self.zip)
            elif hasattr(self, "postal_code") and self.postal_code:
                i18n_address["postal_code"] = str(self.postal_code)
            i18n_address["country_code"] = self._get_country()
            return i18n_address.format_address(i18n_address).replace("\n", line_breaker)  # type: ignore
        output = ""
        if self.city_only is False:
            if (
                (not hasattr(self, "address"))
                and hasattr(self, "street_number")
                and hasattr(self, "street")
            ):
                output += (
                    str(self.street_number) + " " + str(self.street) + line_breaker
                )
            else:
                output += str(self.address) + line_breaker
            the_unit = self.formatted_unit(language=language)
            if the_unit != "":
                output += the_unit + line_breaker
        if hasattr(self, "sublocality_level_1") and self.sublocality_level_1:
            output += str(self.sublocality_level_1) + line_breaker
        output += str(self.city)
        if hasattr(self, "state") and self.state:
            if long_state:
                output += ", " + str(self.state_name())
            else:
                output += ", " + str(self.state)
        if hasattr(self, "zip") and self.zip:
            current_country = (
                self.country if hasattr(self, "country") else get_country()
            )
            if current_country == "US":
                output += " " + str(self.zip).zfill(5)
            else:
                output += " " + str(self.zip)
        elif hasattr(self, "postal_code") and self.postal_code:
            output += " " + str(self.postal_code)
        if (
            show_country is None
            and hasattr(self, "country")
            and self.country
            and get_country() != self.country
        ):
            show_country = True
        if show_country:
            output += line_breaker + country_name(self._get_country())
        return output

    def line_one(
        self,
        language: Optional[str] = None,
        bare: bool = False,
        show_impounded: bool = False,
    ) -> str:
        """Returns the first line of the address, including the unit number if it exists.

        Args:
            language (str, optional): Language for the address format.
            bare (bool): If True, excludes certain formatting elements. Defaults to False.
            show_impounded (bool): If True, shows the address even if impounded. Defaults to False.

        Returns:
            str: The first line of the address.
        """
        if not show_impounded and (hasattr(self, "impounded") and self.impounded):
            return str(self.impounded_output_label)
        if (
            hasattr(self, "has_no_address")
            and self.has_no_address
            and hasattr(self, "has_no_address_explanation")
        ):
            return self.has_no_address_explanation
        if self.city_only:
            return ""
        if (
            (not hasattr(self, "address"))
            and hasattr(self, "street_number")
            and hasattr(self, "street")
        ):
            output = str(self.street_number) + " " + str(self.street)
        else:
            output = str(self.address)
        the_unit = self.formatted_unit(language=language, bare=bare)
        if the_unit != "":
            output += ", " + the_unit
        return output

    def line_two(
        self,
        language: Optional[str] = None,
        long_state: bool = False,
        show_impounded: bool = False,
    ) -> str:
        """Returns the second line of the address, including city, state, and postal code.

        Args:
            language (str, optional): Language for the address format.
            long_state (bool): If True, uses the full state name. Defaults to False.
            show_impounded (bool): If True, shows the address even if impounded. Defaults to False.

        Returns:
            str: The second line of the address.
        """
        if not show_impounded and (hasattr(self, "impounded") and self.impounded):
            return str(self.impounded_output_label)
        output = ""
        # if hasattr(self, 'sublocality') and self.sublocality:
        #    output += str(self.sublocality) + ", "
        if hasattr(self, "sublocality_level_1") and self.sublocality_level_1:
            output += str(self.sublocality_level_1) + ", "
        output += str(self.city)
        if hasattr(self, "state") and self.state:
            if long_state:
                output += ", " + str(self.state_name())
            else:
                output += ", " + str(self.state)
        if hasattr(self, "zip") and self.zip:
            current_country = (
                self.country if hasattr(self, "country") else get_country()
            )
            if current_country == "US":
                output += " " + str(self.zip).zfill(5)
            else:
                output += " " + str(self.zip)
        elif hasattr(self, "postal_code") and self.postal_code:
            output += " " + str(self.postal_code)
        return output

    def on_one_line(
        self,
        include_unit: bool = True,
        omit_default_country: bool = True,
        language: Optional[str] = None,
        show_country: Optional[bool] = None,
        bare: bool = False,
        long_state: bool = False,
        show_impounded: bool = False,
    ) -> str:
        """Returns a one-line formatted address.

        Args:
            include_unit (bool): If True, includes the unit in the formatted address. Defaults to True.
            omit_default_country (bool): If True, doesn't show the Docassemble default country in the formatted address. Defaults to True.
            language (str, optional): Language for the address format.
            show_country (bool, optional): If True, includes the country in the formatted address.
                If None, decides based on the country attribute.
            bare (bool): If True, excludes certain formatting elements. Defaults to False.
            long_state (bool): If True, uses the full state name. Defaults to False.
            show_impounded (bool): If True, shows the address even if impounded. Defaults to False.

        Returns:
            str: The one-line formatted address.
        """
        if not show_impounded and (hasattr(self, "impounded") and self.impounded):
            return str(self.impounded_output_label)
        if (
            hasattr(self, "has_no_address")
            and self.has_no_address
            and hasattr(self, "has_no_address_explanation")
        ):
            return f"{self.has_no_address_explanation}, {self.city} {self.state}"
        output = ""
        if self.city_only is False:
            if (
                (not hasattr(self, "address"))
                and hasattr(self, "street_number")
                and hasattr(self, "street")
            ):
                output += str(self.street_number) + " " + str(self.street)
            else:
                output += str(self.address)
            if include_unit:
                the_unit = self.formatted_unit(language=language, bare=bare)
                if the_unit != "":
                    output += ", " + the_unit
            if output != "":
                output += ", "

        if hasattr(self, "sublocality_level_1") and self.sublocality_level_1:
            if not (
                hasattr(self, "street_number")
                and self.street_number == self.sublocality_level_1
            ):
                output += str(self.sublocality_level_1) + ", "
        output += str(self.city)
        if hasattr(self, "state") and self.state:
            if long_state:
                output += ", " + str(self.state_name())
            else:
                output += ", " + str(self.state)
        if hasattr(self, "zip") and self.zip:
            current_country = (
                self.country if hasattr(self, "country") else get_country()
            )
            if current_country == "US":
                output += " " + str(self.zip).zfill(5)
            else:
                output += " " + str(self.zip)
        elif hasattr(self, "postal_code") and self.postal_code:
            output += " " + str(self.postal_code)
        if (
            show_country is None
            and hasattr(self, "country")
            and self.country
            and ((not omit_default_country) or get_country() != self.country)
        ):
            show_country = True
        if show_country:
            output += ", " + country_name(self._get_country())
        return output

    def normalized_address(self) -> Union[Address, "ALAddress"]:
        """Try geocoding the address, returning the normalized version if successful.

        If geocoding is successful, the method returns the "long" normalized version
        of the address. All methods, such as `my_address.normalized_address().block()`, are
        still available on the returned object. However, note that the returned object will
        be a standard Address object, not an ALAddress object. If geocoding fails, it returns
        the version of the address as entered by the user.

        Warning: currently the normalized address will not be redacted if the address is impounded.

        Returns:
            Union[Address, "ALAddress"]: Normalized address if geocoding is successful, otherwise
                the original address.
        """
        try:
            self.geocode()
        except:
            pass
        if self.was_geocoded_successfully() and hasattr(self, "norm_long"):
            return self.norm_long
        return self

    def state_name(self, country_code: Optional[str] = None) -> str:
        """Returns the full state name based on the state abbreviation.

        If a `country_code` is provided, it will override the country attribute of the Address
        object. Otherwise, the method uses, in order:

        1. The country code associated with the Address object, and then
        2. The country set in the global config for the server.

        Args:
            country_code (str, optional): ISO-3166-1 alpha-2 code to override the country attribute of
                the Address object. For valid codes, refer to:
                https://en.wikipedia.org/wiki/ISO_3166-1_alpha-2#Officially_assigned_code_elements

        Returns:
            str: The full state name corresponding to the state abbreviation. If an error occurs
            or the full name cannot be determined, returns the state abbreviation.
        """
        if country_code:
            return state_name(self.state, country_code=country_code)
        # Do a quick check for a valid ISO country code (alpha-2 only at this time)
        if hasattr(self, "country") and self.country and len(self.country) == 2:
            try:
                return state_name(self.state, country_code=self.country)
            except:
                pass
        try:
            return state_name(
                self.state
            )  # Docassemble falls back to the default set in global config
        except:
            return self.state


class ALAddressList(DAList):
    """A class to store a list of ALAddress objects.

    Extends the DAList class and specifically caters to ALAddress objects.
    It provides methods to initialize the list and get a string representation
    of the list in a formatted manner.
    """

    def init(self, *pargs, **kwargs):
        super(ALAddressList, self).init(*pargs, **kwargs)
        self.object_type = ALAddress

    def __str__(self) -> str:
        """Provide a string representation of the ALAddressList.

        This method returns the addresses in the list formatted in a
        comma-separated manner using the on_one_line method of ALAddress.

        Returns:
            str: Formatted string of all addresses in the list.
        """
        return comma_and_list([item.on_one_line() for item in self])


class ALNameList(DAList):
    """A class to store a list of IndividualName objects.

    Extends the DAList class and is tailored for IndividualName objects.
    """

    def init(self, *pargs, **kwargs):
        super().init(*pargs, **kwargs)
        self.object_type = IndividualName

    def __str__(self) -> str:
        """Provide a string representation of the ALNameList.

        Returns:
            str: Formatted string of all names in the list.
        """
        return comma_list(self)


class ALPeopleList(DAList):
    """Class to store a list of ALIndividual objects, representing people.

    For example, defendants, plaintiffs, or children."""

    def init(self, *pargs, **kwargs):
        super(ALPeopleList, self).init(*pargs, **kwargs)
        self.object_type = ALIndividual

    def names_and_addresses_on_one_line(
        self, comma_string: str = "; ", bare=False
    ) -> str:
        """Provide names and addresses of individuals on one line.

        Args:
            comma_string (str, optional): The string to use between name-address pairs. Defaults to '; '.
            bare (bool, optional): If True, prevents appending the word "Unit" to the unit attribute. Defaults to False.

        Returns:
            str: Formatted string of names followed by addresses.
        """
        return comma_and_list(
            [
                str(person)
                + ", "
                + (
                    person.address.on_one_line(bare=bare)
                    if isinstance(person.address, ALAddress)
                    else str(person.address.on_one_line())
                )
                for person in self
            ],
            comma_string=comma_string,
        )

    def familiar(self) -> str:
        """Provide a list of familiar forms of names of individuals.

        Returns:
            str: Formatted string of familiar names.
        """
        return comma_and_list([person.name.familiar() for person in self])

    def familiar_or(self) -> str:
        """Provide a list of familiar forms of names of individuals separated by 'or'.

        Returns:
            str: Formatted string of familiar names separated by 'or'.
        """
        return comma_and_list(
            [person.name.familiar() for person in self], and_string=word("or")
        )

    def short_list(self, limit: int, truncate_string: str = ", et. al.") -> str:
        """Return a subset of the list, truncated with 'et. al.' if it exceeds a given limit.

        Args:
            limit (int): The maximum number of items to display before truncating.
            truncate_string (str, optional): The string to append when truncating. Defaults to ', et. al.'.

        Returns:
            str: Formatted string of names, truncated if needed.
        """
        if len(self) > limit:
            return comma_and_list(self[:limit]) + truncate_string
        else:
            return comma_and_list(self)

    def full_names(self, comma_string=", ", and_string=word("and")) -> str:
        """Return a formatted list of full names of individuals.

        Args:
            comma_string (str, optional): The string to use between names. Defaults to ','.
            and_string (str, optional): The string to use before the last name in the list. Defaults to 'and'.

        Returns:
            str: Formatted string of full names.
        """
        return comma_and_list(
            [person.name.full(middle="full") for person in self],
            comma_string=comma_string,
            and_string=and_string,
        )


class ALIndividual(Individual):
    """Used to represent an Individual on the assembly line project.

    This class extends the Individual class and adds more tailored attributes and methods
    relevant for the assembly line project. Specifically, it has attributes for previous addresses,
    other addresses, mailing addresses, previous names, aliases, and a preferred name.

    Attributes:
        previous_addresses (ALAddressList): List of previous addresses.
        other_addresses (ALAddressList): List of other addresses.
        mailing_address (ALAddress): Current mailing address.
        service_address (ALAddress): Service address.
        previous_names (ALNameList): List of previous names.
        aliases (ALNameList): List of aliases.
        preferred_name (IndividualName): The preferred name.

    Note:
        Objects as attributes should not be passed directly to the constructor due to
        initialization requirements in the Docassemble framework. See the `init` method.
    """

    previous_addresses: ALAddressList
    other_addresses: ALAddressList
    mailing_address: ALAddress
    service_address: ALAddress
    previous_names: ALNameList
    aliases: ALNameList
    preferred_name: IndividualName

    def init(self, *pargs, **kwargs):
        super(ALIndividual, self).init(*pargs, **kwargs)
        # Initialize the attributes that are themselves objects. Requirement to work with Docassemble
        # See: https://docassemble.org/docs/objects.html#ownclassattributes

        # NOTE: this stops you from passing the address to the constructor
        self.reInitializeAttribute("address", ALAddress)

        if not hasattr(self, "previous_addresses"):
            self.initializeAttribute("previous_addresses", ALAddressList)
        if not hasattr(self, "other_addresses"):
            self.initializeAttribute("other_addresses", ALAddressList)
        if not hasattr(self, "mailing_address"):
            self.initializeAttribute("mailing_address", ALAddress)
        if not hasattr(self, "service_address"):
            self.initializeAttribute("service_address", ALAddress)
        if not hasattr(self, "previous_names"):
            self.initializeAttribute("previous_names", ALNameList)
        if not hasattr(self, "aliases"):
            self.initializeAttribute("aliases", ALNameList)
        if not hasattr(self, "preferred_name"):
            self.initializeAttribute("preferred_name", IndividualName)

    def signature_if_final(self, i: str) -> Union[DAFile, str]:
        """Returns the individual's signature if `i` is "final", which usually means we are assembling the final version of the document (as opposed to a preview).

        Args:
            i (str): The condition which, if set to "final", returns the signature.

        Returns:
            Union[DAFile, str]: The signature if the condition is met, otherwise an empty string.
        """
        if i == "final":
            return self.signature
        else:
            return ""

    def phone_numbers(
        self, country: Optional[str] = None, show_impounded: bool = False
    ) -> str:
        """Fetches and formats the phone numbers of the individual.

        Args:
            country (str, optional): The country for phone number formatting. Defaults to the country of the docassemble server.
            show_impounded (bool): If True, shows the phone numbers even if impounded. Defaults to False.

        Returns:
            str: Formatted string of phone numbers.
        """
        nums = []
        if hasattr(self, "mobile_number") and self.mobile_number:
            fmt_number = None
            try:
                fmt_number = phone_number_formatted(self.mobile_number, country=country)
            except:
                fmt_number = None
            if fmt_number:
                nums.append({fmt_number: "cell"})
            else:
                nums.append({self.mobile_number: "cell"})
        if hasattr(self, "phone_number") and self.phone_number:
            fmt_number = None
            try:
                fmt_number = phone_number_formatted(self.phone_number, country=country)
            except:
                fmt_number = None
            if fmt_number:
                nums.append({fmt_number: "other"})
            else:
                nums.append({self.phone_number: "other"})
        if len(nums) < 1:
            return ""
        # Check for impounded phone number
        elif not show_impounded and (
            hasattr(self, "phone_impounded") and self.phone_impounded
        ):
            return str(self.impounded_phone_output_label)
        elif len(nums) > 1:
            return comma_list(
                [f"{list(num.keys())[0]} ({list(num.values())[0]})" for num in nums]
            )
        elif len(nums):
            return list(nums[0].keys())[0]

        assert False  # We should never get here, no default return is necessary

    def contact_methods(self) -> str:
        """Generates a formatted string of all provided contact methods.

        Returns:
            str: A formatted string indicating the available methods to contact the individual.
        """
        methods = []
        if self.phone_numbers():
            methods.append({self.phone_numbers(): str(self.phone_number_contact_label)})
        if hasattr(self, "email") and self.email:
            methods.append({self.email: str(self.email_contact_label)})
        if hasattr(self, "other_contact_method") and self.other_contact_method:
            methods.append(
                {self.other_contact_method: str(self.other_contact_method_label)}
            )

        return comma_and_list(
            [
                list(method.values())[0] + " " + list(method.keys())[0]
                for method in methods
                if len(method)
            ],
            and_string=word("or"),
        )

    def merge_letters(self, new_letters: str) -> None:
        """If the Individual has a child_letters attribute, add the new letters to the existing list

        Avoid using. Only used in 209A.

        Args:
            new_letters (str): The new letters to add to the existing list of letters
        """
        # TODO: move to 209A package
        if hasattr(self, "child_letters"):
            self.child_letters: str = filter_letters([new_letters, self.child_letters])
        else:
            self.child_letters = filter_letters(new_letters)

    def formatted_age(self) -> str:
        """Calculates and formats the age of the individual based on their birthdate.

        Returns:
            str: Formatted age string that shows the most relevant time unit; for example, if under 2 years, it will return "X months".
        """
        dd = date_difference(self.birthdate)
        if dd.years >= 2:
            return "%d years" % (int(dd.years),)
        if dd.weeks > 12:
            return "%d months" % (int(dd.years * 12.0),)
        if dd.weeks > 2:
            return "%d weeks" % (int(dd.weeks),)
        return "%d days" % (int(dd.days),)

    def normalized_address(self) -> Union[Address, ALAddress]:
        """Fetches the normalized version of the address.

        Returns:
            Union[Address, ALAddress]: The normalized address object.
        """
        return self.address.normalized_address()

    # This design helps us translate the prompts for common fields just once
    def name_fields(
        self,
        person_or_business: str = "person",
        show_suffix: bool = True,
        show_title: bool = False,
        title_options: Optional[List[str]] = None,
        show_if: Union[str, Dict[str, str], None] = None,
    ) -> List[Dict[str, str]]:
        """
        Generates suitable field prompts for a name based on the type of entity (person or business)
        and other provided parameters.

        Args:
            person_or_business (str, optional): Specifies the entity type. It can either be "person" or "business".
                Default is "person".
            show_suffix (bool, optional): Determines if the name's suffix (e.g., Jr., Sr.) should be included in the prompts.
                Default is True.
            show_title: (bool, optional): Determines if the name's title (e.g., Mr., Ms.) should be included in the prompts.
                Default is False.
            title_options (List[str], optional): A list of title options to use in the prompts. Default is defined as a list
                of common titles in English-speaking countries.
            show_if (Union[str, Dict[str, str], None], optional): Condition to determine which fields to show.
                It can be a string, a dictionary with conditions, or None. Default is None.

        Returns:
            List[Dict[str, str]]: A list of dictionaries where each dictionary contains field prompt details.

        Note:
            If `person_or_business` is set to None, the method will offer the end user a choice
            and will set appropriate "show ifs" conditions for each type.
        """
        if not title_options:
            title_options = [
                "Mr.",
                "Mrs.",
                "Miss",
                "Ms.",
                "Mx.",
                "Dr.",
                "Prof.",
                "Hon.",
                "Rev.",
                "Sir",
                "Lord",
                "Lady",
                "Dame",
                "Maj.",
                "Gen.",
                "Capt.",
                "Lt.",
                "Sgt.",
                "Fr.",
                "Sr.",
            ]
        if person_or_business == "person":
            fields = [
                {
                    "label": str(self.first_name_label),
                    "field": self.attr_name("name.first"),
                },
                {
                    "label": str(self.middle_name_label),
                    "field": self.attr_name("name.middle"),
                    "required": False,
                },
                {
                    "label": str(self.last_name_label),
                    "field": self.attr_name("name.last"),
                },
            ]
            if show_suffix:
                fields.append(
                    {
                        "label": str(self.suffix_label),
                        "field": self.attr_name("name.suffix"),
                        "choices": name_suffix(),
                        "required": False,
                    }
                )
            if show_title:
                fields.insert(
                    0,
                    {
                        "label": str(self.name_title_label),
                        "field": self.attr_name("name.title"),
                        "choices": title_options,
                        "required": False,
                    },
                )
            if show_if:
                for field in fields:
                    field["show if"] = show_if
            return fields
        elif person_or_business == "business":
            # Note: we don't make use of the name.text field for simplicity
            # TODO: this could be reconsidered`, but name.text tends to lead to developer error
            fields = [
                {
                    "label": str(self.business_name_label),
                    "field": self.attr_name("name.first"),
                }
            ]
            if show_if:
                fields[0]["show if"] = show_if
            return fields
        else:
            # Note: the labels are template block objects: if they are keys,
            # they should be converted to strings first
            show_if_indiv = {
                "variable": self.attr_name("person_type"),
                "is": "ALIndividual",
            }
            show_if_business = {
                "variable": self.attr_name("person_type"),
                "is": "business",
            }
            fields = [
                {
                    "label": str(self.person_type_label),
                    "field": self.attr_name("person_type"),
                    "choices": [
                        {str(self.individual_choice_label): "ALIndividual"},
                        {str(self.business_choice_label): "business"},
                    ],
                    "input type": "radio",
                    "required": True,
                },
                # Individual questions
                {
                    "label": str(self.first_name_label),
                    "field": self.attr_name("name.first"),
                    "show if": show_if_indiv,
                },
                {
                    "label": str(self.middle_name_label),
                    "field": self.attr_name("name.middle"),
                    "required": False,
                    "show if": show_if_indiv,
                },
                {
                    "label": str(self.last_name_label),
                    "field": self.attr_name("name.last"),
                    "show if": show_if_indiv,
                },
            ]
            if show_if:
                fields[0]["show if"] = show_if

            if show_suffix:
                fields.append(
                    {
                        "label": str(self.suffix_label),
                        "field": self.attr_name("name.suffix"),
                        "choices": name_suffix(),
                        "required": False,
                        "show if": show_if_indiv,
                    }
                )
            fields.append(
                # Business names
                {
                    "label": str(self.business_name_label),
                    "field": self.attr_name("name.first"),
                    "show if": show_if_business,
                }
            )
            return fields

    def address_fields(
        self,
        country_code: str = "US",
        default_state: Optional[str] = None,
        show_country: bool = False,
        show_county: bool = False,
        show_if: Union[str, Dict[str, str], None] = None,
        allow_no_address: bool = False,
        ask_if_impounded: bool = False,
    ) -> List[Dict[str, str]]:
        """
        Generate field prompts for capturing an address.

        Args:
            country_code (str): The default country for the address. Defaults to "US".
            default_state (Optional[str]): Default state if applicable. Defaults to None.
            show_country (bool): Whether to display the country field. Defaults to False.
            show_county (bool): Whether to display the county field. Defaults to False.
            show_if (Union[str, Dict[str, str], None]): Condition to determine if the field should be shown. Defaults to None.
            allow_no_address (bool): Whether to permit entries with no address. Defaults to False.
            ask_if_impounded (bool): Whether to ask if the address is impounded. Defaults to False.

        Returns:
            List[Dict[str, str]]: A list of dictionaries with field prompts for addresses.
        """
        # TODO make this more flexible to work w/ homeless individuals and
        # international addresses
        return self.address.address_fields(
            country_code=country_code,
            default_state=default_state,
            show_country=show_country,
            show_county=show_county,
            show_if=show_if,
            allow_no_address=allow_no_address,
            ask_if_impounded=ask_if_impounded,
        )

    def gender_fields(
        self, show_help=False, show_if: Union[str, Dict[str, str], None] = None
    ) -> List[Dict[str, str]]:
        """
        Generate fields for capturing gender information, including a
        self-described option.

        Args:
            show_help (bool): Whether to show additional help text. Defaults to False.
            show_if (Union[str, Dict[str, str], None]): Condition to determine if the field should be shown. Defaults to None.

        Returns:
            List[Dict[str, str]]: A list of dictionaries with field prompts for gender.
        """
        choices = [
            {str(self.gender_female_label): "female"},
            {str(self.gender_male_label): "male"},
            {str(self.gender_nonbinary_label): "nonbinary"},
            {str(self.gender_prefer_not_to_say_label): "prefer-not-to-say"},
            {str(self.gender_prefer_self_described_label): "self-described"},
            {str(self.gender_unknown_label): "unknown"},
        ]
        self_described_input = {
            "label": str(self.gender_self_described_label),
            "field": self.attr_name("gender"),
            "show if": {"variable": self.attr_name("gender"), "is": "self-described"},
        }
        fields = [
            {
                "label": str(self.gender_label),
                "field": self.attr_name("gender"),
                "choices": choices,
            },
            self_described_input,
        ]

        if show_help:
            fields[0]["help"] = str(self.gender_help_text)
        if show_if:
            fields[0]["show if"] = show_if

        return fields

    def pronoun_fields(
        self,
        show_help=False,
        show_if: Union[str, Dict[str, str], None] = None,
        required: bool = False,
        shuffle: bool = False,
        show_unknown: Optional[Union[Literal["guess"], bool]] = "guess",
    ) -> List[Dict[str, str]]:
        """
        Generate fields for capturing pronoun information.

        Args:
            show_help (bool): Whether to show additional help text. Defaults to False.
            show_if (Union[str, Dict[str, str], None]): Condition to determine if the field should be shown. Defaults to None.
            required (bool): Whether the field is required. Defaults to False.
            shuffle (bool): Whether to shuffle the order of pronouns. Defaults to False.
            show_unknown (Union[Literal["guess"], bool]): Whether to show an "unknown" option. Can be "guess", True, or False. Defaults to "guess".

        Returns:
            List[Dict[str, str]]: A list of dictionaries with field prompts for pronouns.
        """
        shuffled_choices = [
            {str(self.pronoun_she_label): "she/her/hers"},
            {str(self.pronoun_he_label): "he/him/his"},
            {str(self.pronoun_they_label): "they/them/theirs"},
            {str(self.pronoun_zir_label): "ze/zir/zirs"},
        ]
        if shuffle:
            random.shuffle(shuffled_choices)
        final_choices = [
            {str(self.pronoun_prefer_self_described_label): "self-described"},
        ]
        if show_unknown == True or (
            show_unknown == "guess" and self.instanceName != "users[0]"
        ):
            final_choices.append({str(self.pronoun_unknown_label): "unknown"})
        self_described_input = {
            "label": str(self.pronoun_self_described_label),
            "field": self.attr_name("pronouns_self_described"),
            "show if": self.attr_name("pronouns['self-described']"),
        }
        fields = [
            {
                "label": str(self.pronouns_label),
                "field": self.attr_name("pronouns"),
                "datatype": "checkboxes",
                "choices": shuffled_choices + final_choices,
                "none of the above": str(self.pronoun_prefer_not_to_say_label),
                "required": required,
            },
            self_described_input,
        ]

        if show_help:
            fields[0]["help"] = str(self.pronouns_help_text)
        if show_if:
            fields[0]["show if"] = show_if

        return fields

    def get_pronouns(self) -> set:
        """
        Retrieve a set of the individual's pronouns.

        If the individual has selected the "self-described" option, it will use their custom input.

        Can be formatted however the author likes.

        Returns:
            set: A set of strings representing the individual's pronouns.
        """
        if hasattr(self, "pronouns") and isinstance(self.pronouns, str):
            return {self.pronouns}
        if self.pronouns.all_false():
            return {str(self.pronoun_prefer_not_to_say_label)}
        pronouns = set(self.pronouns.true_values()) - {"self-described"}
        if self.pronouns.get("self-described"):
            pronouns = pronouns.union(self.pronouns_self_described.splitlines())
        return pronouns

    def list_pronouns(self) -> str:
        """
        Retrieve a formatted string of the individual's pronouns, arranged with
        the comma_list() function.

        Returns:
            str: A formatted string of the individual's pronouns.
        """
        return comma_list(sorted(self.get_pronouns()))

    def language_fields(
        self,
        choices: Optional[List[Dict[str, str]]] = None,
        style: str = "radio",
        show_if: Union[str, Dict[str, str], None] = None,
    ) -> List[Dict[str, str]]:
        """
        Generate fields for capturing language preferences.

        Args:
            choices (Optional[List[Dict[str, str]]]): A list of language choices. Defaults to None.
            style (str): The display style of choices. Defaults to "radio".
            show_if (Union[str, Dict[str, str], None]): Condition to determine if the field should be shown. Defaults to None.

        Returns:
            List[Dict[str, str]]: A list of dictionaries with field prompts for language preferences.
        """
        if not choices:
            choices = [
                {"English": "en"},
                {"Spanish": "es"},
                {"Other": "other"},
            ]
        other = {
            "label": str(self.language_other_label),
            "field": self.attr_name("language_other"),
            "show if": {"variable": self.attr_name("language"), "is": "other"},
        }
        fields = [
            {
                "label": str(self.language_label),
                "field": self.attr_name("language"),
                "choices": choices,
            },
            other,
        ]
        if style == "radio":
            fields[0]["input type"] = "radio"
        if show_if:
            fields[0]["show if"] = show_if
        return fields

    def language_name(self) -> str:
        """
        Get the human-readable version of the individual's selected language.

        Returns:
            str: The human-readable version of the language. If 'other' is selected,
            it returns the value in `language_other`. Otherwise, it uses the
            `language_name` function.
        """
        if self.language == "other":
            return self.language_other
        else:
            return language_name(self.language)

    @property
    def gender_male(self) -> bool:
        """
        Returns True only if the gender is male.

        Used to assist with checkbox filling in PDFs with "skip undefined"
        turned on.
        """
        return self.gender.lower() == "male"

    @property
    def gender_female(self) -> bool:
        """
        Returns True only if the gender is female.

        Used to assist with checkbox filling in PDFs with "skip undefined"
        turned on.
        """
        return self.gender.lower() == "female"

    @property
    def gender_other(self) -> bool:
        """
        Returns True only if the gender is not male or female.

        Used to assist with checkbox filling in PDFs with "skip undefined"
        turned on.
        """
        return (self.gender != "male") and (self.gender != "female")

    @property
    def gender_nonbinary(self) -> bool:
        """
        Returns True only if the gender is nonbinary.

        Used to assist with checkbox filling in PDFs with "skip undefined"
        turned on.
        """
        return self.gender.lower() == "nonbinary"

    @property
    def gender_unknown(self) -> bool:
        """
        Returns True only if the gender is unknown.

        Used to assist with checkbox filling in PDFs with "skip undefined"
        turned on.
        """
        return self.gender.lower() == "unknown"

    @property
    def gender_undisclosed(self) -> bool:
        """
        Returns True only if the gender is not disclosed ("prefer-not-to-say")

        Used to assist with checkbox filling in PDFs with "skip undefined"
        turned on.
        """
        return self.gender.lower() == "prefer-not-to-say"

    @property
    def gender_self_described(self) -> bool:
        """
        Returns True only if the gender is self described.

        Used to assist with checkbox filling in PDFs with "skip undefined"
        turned on.
        """
        return not (
            self.gender
            in ["prefer-not-to-say", "male", "female", "unknown", "nonbinary"]
        )

    def contact_fields(self) -> None:
        """
        Return field prompts for other contact info
        """
        pass

    @property
    def initials(self) -> str:
        """
        Returns the initials of the individual as a string.

        For example, "Quinten K Steenhuis" would return "QKS".
        """
        return f"{self.name.first[:1]}{self.name.middle[:1] if hasattr(self.name,'middle') else ''}{self.name.last[:1] if hasattr(self.name, 'last') else ''}"

    def address_block(
        self,
        language=None,
        international=False,
        show_country=False,
        bare=False,
        show_impounded=False,
    ) -> str:
        """
        Generate a formatted address block for mailings.

        Args:
            language (Optional): The language in which the address is written.
            international (bool): If True, format for international mailing. Defaults to False.
            show_country (bool): If True, include the country in the address. Defaults to False.
            bare (bool): If True, produce the address without additional formatting. Defaults to False.
            show_impounded (bool): If True, show the address even if it is impounded. Defaults to False.

        Returns:
            str: The formatted address block.
        """
        if this_thread.evaluation_context == "docx":
            if isinstance(self.address, ALAddress):
                return self.address.block(
                    language=language,
                    international=international,
                    show_country=show_country,
                    bare=bare,
                    show_impounded=show_impounded,
                )
            else:
                # bare parameter is ignored for plain Address objects
                return self.address.block(
                    language=language,
                    international=international,
                    show_country=show_country,
                )
        else:
            if isinstance(self.address, ALAddress):
                return (
                    "[FLUSHLEFT] "
                    + self.name.full()
                    + " [NEWLINE] "
                    + self.address.block(
                        language=language,
                        international=international,
                        show_country=show_country,
                        bare=bare,
                        show_impounded=show_impounded,
                    )
                )
            else:
                return (
                    "[FLUSHLEFT] "
                    + self.name.full()
                    + " [NEWLINE] "
                    + self.address.block(
                        language=language,
                        international=international,
                        show_country=show_country,
                    )
                )

    def pronoun(self, **kwargs) -> str:
        """Returns an objective pronoun as appropriate, based on attributes.

        The pronoun could be "you," "her," "him," "it," or "them". It depends
        on the `gender` and `person_type` attributes and whether the individual
        is the current user.

        If the user selected specific pronouns, they take priority over
        gender (only if they chose a pronoun from the list)

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            str: The appropriate pronoun.
        """
        # Developers can do funky things; be a little flexible
        if hasattr(self, "pronouns") and isinstance(self.pronouns, str):
            pronouns = DADict(elements={self.pronouns.lower(): True})
        else:
            pronouns = self.pronouns

        if self == this_thread.global_vars.user:
            output = word("you", **kwargs)
        elif (
            hasattr(self, "pronouns")
            and isinstance(pronouns, DADict)
            and len(pronouns.true_values()) == 1
            and (
                (
                    pronouns.true_values()[0]
                    in ["she/her/hers", "he/him/his", "they/them/theirs", "ze/zir/zirs"]
                )
                or (
                    pronouns.get("self-described")
                    and has_parsable_pronouns(self.pronouns_self_described)
                )
            )
        ):
            if pronouns.get("she/her/hers"):
                output = word("her", **kwargs)
            elif pronouns.get("he/him/his"):
                output = word("him", **kwargs)
            elif pronouns.get("they/them/theirs"):
                output = word("them", **kwargs)
            elif pronouns.get("ze/zir/zirs"):
                output = word("zir", **kwargs)
            elif pronouns.get("self-described"):
                output = parse_custom_pronouns(self.pronouns_self_described)["o"]
        elif hasattr(self, "person_type") and self.person_type in [
            "business",
            "organization",
        ]:
            output = word("it", **kwargs)
        elif self.gender.lower() == "female":
            output = word("her", **kwargs)
        elif self.gender.lower() == "male":
            output = word("him", **kwargs)
        else:
            output = word("them", **kwargs)
        if "capitalize" in kwargs and kwargs["capitalize"]:
            return capitalize(output)
        return output

    def pronoun_objective(self, **kwargs) -> str:
        """Returns the same pronoun as the `pronoun()` method.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            str: The appropriate objective pronoun.
        """
        return self.pronoun(**kwargs)

    def pronoun_possessive(self, target, **kwargs) -> str:
        """Returns a possessive pronoun and a target word, based on attributes.

        Given a target word, the function returns "{pronoun} {target}". The pronoun could be
        "her," "his," "its," or "their". It depends on the `gender` and `person_type` attributes
        and whether the individual is the current user.

        Args:
            target (str): The target word to follow the pronoun.
            **kwargs: Additional keyword arguments.

        Returns:
            str: The appropriate possessive phrase.
        """
        if hasattr(self, "pronouns") and isinstance(self.pronouns, str):
            pronouns = DADict(elements={self.pronouns.lower(): True})
        else:
            pronouns = self.pronouns

        if self == this_thread.global_vars.user and (
            "thirdperson" not in kwargs or not kwargs["thirdperson"]
        ):
            output = your(target, **kwargs)
        elif (
            hasattr(self, "pronouns")
            and isinstance(pronouns, DADict)
            and len(pronouns.true_values()) == 1
            and (
                (
                    pronouns.true_values()[0]
                    in ["she/her/hers", "he/him/his", "they/them/theirs", "ze/zir/zirs"]
                )
                or (
                    pronouns.get("self-described")
                    and has_parsable_pronouns(self.pronouns_self_described)
                )
            )
        ):
            if pronouns.get("she/her/hers"):
                output = her(target, **kwargs)
            elif pronouns.get("he/him/his"):
                output = his(target, **kwargs)
            elif pronouns.get("they/them/theirs"):
                output = their(target, **kwargs)
            elif pronouns.get("ze/zir/zirs"):
                output = word("zir", **kwargs) + " " + target
            elif pronouns.get("self-described"):
                output = (
                    parse_custom_pronouns(self.pronouns_self_described)["p"]
                    + " "
                    + target
                )
        elif hasattr(self, "person_type") and self.person_type in [
            "business",
            "organization",
        ]:
            output = its(target, **kwargs)
        elif self.gender.lower() == "female":
            output = her(target, **kwargs)
        elif self.gender.lower() == "male":
            output = his(target, **kwargs)
        else:
            output = their(target, **kwargs)
        if "capitalize" in kwargs and kwargs["capitalize"]:
            return capitalize(output)
        return output

    def pronoun_subjective(self, **kwargs) -> str:
        """Returns a subjective pronoun, based on attributes.

        The pronoun could be "you," "she," "he," "it," or "they". It depends
        on the `gender` and `person_type` attributes and whether the individual
        is the current user.

        Args:
            **kwargs: Additional keyword arguments.

        Returns:
            str: The appropriate subjective pronoun.
        """
        if hasattr(self, "pronouns") and isinstance(self.pronouns, str):
            pronouns = DADict(elements={self.pronouns.lower(): True})
        else:
            pronouns = self.pronouns

        if self == this_thread.global_vars.user and (
            "thirdperson" not in kwargs or not kwargs["thirdperson"]
        ):
            output = word("you", **kwargs)
        elif (
            hasattr(self, "pronouns")
            and isinstance(pronouns, DADict)
            and len(pronouns.true_values()) == 1
            and (
                (
                    pronouns.true_values()[0]
                    in ["she/her/hers", "he/him/his", "they/them/theirs", "ze/zir/zirs"]
                )
                or (
                    pronouns.get("self-described")
                    and has_parsable_pronouns(self.pronouns_self_described)
                )
            )
        ):
            if pronouns.get("she/her/hers"):
                output = word("she", **kwargs)
            elif pronouns.get("he/him/his"):
                output = word("he", **kwargs)
            elif pronouns.get("they/them/theirs"):
                output = word("they", **kwargs)
            elif pronouns.get("ze/zir/zirs"):
                output = word("ze", **kwargs)
            elif pronouns.get("self-described"):
                output = parse_custom_pronouns(self.pronouns_self_described)["s"]
        elif hasattr(self, "person_type") and self.person_type in [
            "business",
            "organization",
        ]:
            output = word("it", **kwargs)
        elif self.gender.lower() == "female":
            output = word("she", **kwargs)
        elif self.gender.lower() == "male":
            output = word("he", **kwargs)
        else:
            output = word("they", **kwargs)
        if "capitalize" in kwargs and kwargs["capitalize"]:
            return capitalize(output)
        return output

    def name_full(self) -> str:
        """Returns the individual's full name.

        Returns:
            str: The individual's full name.
        """
        return self.name.full(middle="full")

    def name_initials(self) -> str:
        """
        Returns the individual's name with the middle name as an initial.
        Equivalent to `name.full(middle="initial")`, which is also the default.
        Defined only to make it possible to be explicit about the name form.

        Returns:
            str: The individual's name with the middle name as an initial.
        """
        return self.name.full(middle="initial")

    def name_short(self) -> str:
        """
        Returns the individual's name without any middle name.

        Equivalent to self.name.firstlast()

        Returns:
            str: The individual'
        """
        return self.name.firstlast()


# (DANav isn't in public DA API, but currently in functions.py)
def section_links(nav) -> List[str]:  # type: ignore
    """Returns a list of clickable navigation links without animation.

    Args:
        nav: The navigation object.

    Returns:
        List[str]: A list of clickable navigation links without animation.
    """
    sections = nav.get_sections()
    section_link = []
    for section in sections:
        for key in section:
            section_link.append("[" + section[key] + "](" + url_action(key) + ")")

    return section_link


########################################################
# Subject-specific classes


class Landlord(ALIndividual):
    pass


class Tenant(ALIndividual):
    pass


class HousingAuthority(Landlord):
    pass


class Applicant(Tenant):
    pass


class Abuser(ALIndividual):
    pass


class Survivor(ALIndividual):
    pass


########################################################
# Compatibility layer to help with migration

# TODO: consider removing after packages migrated


class VCIndividual(ALIndividual):
    pass


class AddressList(ALAddressList):
    pass


class PeopleList(ALPeopleList):
    pass


########################################################
# Miscellaneous functions needed for baseline questions
# These could go in toolbox but keeping here to reduce packages
# needed for baseline running.


def will_send_to_real_court() -> bool:
    """
    For legacy email to court forms, this checks to see if the form
    is being run on the dev, test, or production server.

    The text "dev" or "test" needs to be in the URL root in the DA config: can change in `/config`.

    Returns:
        bool: True if the form is being run on the dev, test, or production server.
    """
    return not (
        get_config("debug")
        or "dev" in get_config("url root")
        or "test" in get_config("url root")
        or "localhost" in get_config("url root")
    )


# TODO: move to 209A package
# This one is only used for 209A--should move there along with the combined_letters() method
def filter_letters(letter_strings: Union[List[str], str]) -> str:
    """Used to take a list of letters like ["A","ABC","AB"] and filter out any duplicate letters.

    Avoid using, this is created for 209A.

    Args:
        letter_strings (Union[List[str], str]): A list of letters.

    Returns:
        str: A string of unique letters.
    """
    # There is probably a cute one liner, but this is easy to follow and
    # probably same speed
    unique_letters = set()
    if isinstance(letter_strings, str):
        letter_strings = [letter_strings]
    for string in letter_strings:
        if string:  # Catch possible None values
            for letter in string:
                unique_letters.add(letter)
    try:
        retval = "".join(sorted(unique_letters))
    except:
        retval = ""
    return retval


# Note: removed "combined_locations" because it is too tightly coupled to MACourts.py right now


def fa_icon(
    icon: str, color: str = "primary", color_css: Optional[str] = None, size: str = "sm"
) -> str:
    """
    Return HTML for a font-awesome icon of the specified size and color. You can reference
    a CSS variable (such as Bootstrap theme color) or a true CSS color reference, such as 'blue' or
    '#DDDDDD'. Defaults to Bootstrap theme color "primary".

    Args:
        icon (str): The name of the icon to use. See https://fontawesome.com/icons for a list of icons.
        color (str): The color of the icon. Defaults to "primary".
        color_css (Optional[str]): A CSS variable or color reference. Defaults to None.
        size (str): The size of the icon. Defaults to "sm".

    Returns:
        str: HTML for the icon.
    """
    if not color and not color_css:
        return ":" + icon + ":"  # Default to letting Docassemble handle it
    elif color_css:
        return (
            '<i class="fa fa-'
            + icon
            + " fa-"
            + size
            + '" style="color:'
            + color_css
            + ';"></i>'
        )
    else:
        return (
            '<i class="fa fa-'
            + icon
            + " fa-"
            + size
            + '" style="color:var(--'
            + color
            + ');"></i>'
        )


def is_sms_enabled() -> bool:
    """Checks if SMS (Twilio) is enabled on the server. Does not verify that it works.

    See https://docassemble.org/docs/config.html#twilio for more info.

    Returns:
        bool: True if there is a non-empty Twilio config on the server, False otherwise
    """
    twilio_config = get_config("twilio")
    if isinstance(twilio_config, list):
        to_check = twilio_config[0]
    elif isinstance(twilio_config, dict):
        to_check = twilio_config
    else:
        return False

    return bool(
        to_check.get("sms")
        and to_check.get("account sid")
        and to_check.get("auth token")
        and to_check.get("number")
    )


def is_phone_or_email(text: str) -> bool:
    """
    Returns True if the string is either a valid phone number or a valid email address.
    If SMS is not enabled on the server (through the Twilio config), only accepts emails.
    Email validation is extremely minimal--just checks for an @ sign between two non-zero length
    strings.

    Args:
        text (str): The string to check.

    Returns:
        bool: True if the string is either a valid phone number or a valid email address.

    Raises:
        DAValidationError if the string is neither a valid phone number nor a valid email address.
    """
    sms_enabled = is_sms_enabled()
    if re.match("\S+@\S+", text) or (sms_enabled and phone_number_is_valid(text)):
        return True
    else:
        if sms_enabled:
            validation_error("Enter a valid phone number or email address")
        else:
            validation_error("Enter a valid email address")
        assert False, "unreachable"


def github_modified_date(
    github_user: str, github_repo_name: str, auth=None
) -> Union[DADateTime, None]:
    """
    Returns the date that the given GitHub repository was modified or None if API call fails.

    Will check for the presence of credentials in the configuration labeled "github issues"
    in this format:

    github issues:
      username: YOUR_GITHUB_USERNAME
      token: YOUR_GITHUB_PRIVATE_TOKEN

    If those credentials aren't found, it will then search for credentials in this format (deprecated):

    github readonly:
      username: YOUR_GITHUB_USERNAME
      password: YOUR_GITHUB_PRIVATE_TOKEN
      type: basic

    If no valid auth information is in the configuration, it will fall back to anonymous authentication.
    The GitHub API is rate-limited to 60 anonymous API queries/hour.

    Args:
        github_user (str): The GitHub username of the repository owner.
        github_repo_name (str): The name of the repository.
        auth (Optional[dict]): A dictionary containing authentication information. Defaults to None.

    Returns:
        Union[DADateTime, None]: The date that the given GitHub repository was modified or None if API call fails.
    """
    if not auth:
        issue_config = get_config("github issues")
        if issue_config:
            auth = {
                "username": issue_config.get("username"),
                "password": issue_config.get("token"),
                "type": "basic",
            }
        else:
            auth = get_config("github readonly")
    github_readonly_web = DAWeb(base_url="https://api.github.com")
    res = github_readonly_web.get(
        "repos/" + github_user + "/" + github_repo_name,
        auth=auth,
    )
    if res and res.get("pushed_at"):
        return as_datetime(res.get("pushed_at"))
    else:
        return None


# TODO(qs): remove if https://github.com/jhpyle/docassemble/pull/520 is merged
def language_name(language_code: str) -> str:
    """Given a 2 digit language code abbreviation, returns the full
    name of the language. The language name will be passed through the `word()`
    function.

    Args:
        language_code (str): A 2 digit language code abbreviation.

    Returns:
        str: The full name of the language.
    """
    ensure_definition(language_code)
    try:
        if len(language_code) == 2:
            return word(pycountry.languages.get(alpha_2=language_code).name)
        else:
            return word(pycountry.languages.get(alpha_3=language_code).name)
    except:
        return language_code


def safe_states_list(country_code: str) -> List[Dict[str, str]]:
    """Wrapper around states_list that doesn't error if passed
    an invalid country_code (e.g., a country name spelled out)

    Args:
        country_code (str): A 2 digit country code abbreviation.

    Returns:
        List[Dict[str, str]]: A list of dictionaries with field prompts for states.
    """
    try:
        return states_list(country_code=country_code)
    except:
        return states_list()


def has_parsable_pronouns(pronouns: str) -> bool:
    """
    Returns True if the pronouns string can be parsed into a dictionary of pronouns.

    Args:
        pronouns: a string of pronouns in the format "objective/subjective/possessive"

    Returns:
        True if the pronouns string can be parsed into a dictionary of pronouns, False otherwise
    """
    try:
        parse_custom_pronouns(pronouns)
        return True
    except:
        return False


def parse_custom_pronouns(pronouns: str) -> Dict[str, str]:
    """
    Parses a custom pronoun string into a dictionary of pronouns.

    Args:
        pronouns: a string of pronouns in the format "objective/subjective/possessive"

    Returns:
        a dictionary of pronouns in the format {"o": objective, "s": subjective, "p": possessive}
    """
    # test for presence of either 2 or 3 /'s
    if not (2 <= pronouns.count("/") <= 3):
        raise Exception("Pronouns must contain either 2 or 3 slashes.")

    pronoun_list = pronouns.split("/")
    # entry 1 is objective, entry 2 is subjective, entry 3 is possessive, entry 4 is possessive pronoun (unused)
    return {
        "o": pronoun_list[0].lower(),
        "s": pronoun_list[1].lower(),
        "p": pronoun_list[2].lower(),
    }


def get_visible_al_nav_items(
    nav_items: List[Union[str, dict]]
) -> List[Union[str, dict]]:
    """
    Processes a list of nav items and returns only the ones that are not hidden.
    Can be used to control the visible nav items in a more declarative way while keeping
    the navigation dynamic.

    Expects a list like this:

    data = [
        {"key": "value", "hidden": True},
        "top level item",
        {"key2": [{"subkey": "subvalue", "hidden": False}, {"subkey": "subvalue2", "hidden": True}]},
    ]

    Args:
        nav_items: a list of nav items

    Returns:
        a list of nav items with hidden items removed
    """
    new_list: List[Union[str, dict]] = []

    for item in nav_items:
        # For strings at top level
        if isinstance(item, str):
            new_list.append(item)
            continue

        # For dictionaries at top level
        item_copy = deepcopy(item)
        if not str(item_copy.pop("hidden", "False")).lower() == "true":  # if not hidden
            for key, val in item_copy.items():
                if isinstance(val, list):  # if value of a key is a list
                    new_sublist: List[Union[str, dict]] = []
                    for subitem in val:
                        # Add subitem strings as-is
                        if isinstance(subitem, str):
                            new_sublist.append(subitem)
                        # Add dictionaries if not hidden
                        elif (
                            isinstance(subitem, dict)
                            and not str(subitem.pop("hidden", "False")).lower()
                            == "true"
                        ):
                            new_sublist.append(subitem)
                    item_copy[key] = new_sublist
            new_list.append(item_copy)
        continue

    return new_list



import unittest
from .al_general import ALIndividual, ALAddress, get_visible_al_nav_items
from unittest.mock import Mock
from docassemble.base.util import DADict, DAAttributeError


class test_aladdress(unittest.TestCase):
    def setUp(self):
        self.addr = ALAddress()

    def test_empty_is_empty(self):
        addr = ALAddress(address="", city="")
        self.assertEqual(addr.on_one_line(), "")

    def test_no_unit_floor_room(self):
        self.assertFalse(hasattr(self.addr, "unit"))
        self.assertFalse(hasattr(self.addr, "floor"))
        self.assertFalse(hasattr(self.addr, "room"))
        self.assertEqual(self.addr.formatted_unit(), "")

    def test_require_unit_exception(self):
        self.assertFalse(hasattr(self.addr, "unit"))
        with self.assertRaises(AttributeError):
            self.addr.formatted_unit(require=True)

    def test_unit_isnumeric(self):
        self.addr.unit = "123"
        self.assertEqual(self.addr.formatted_unit(), "Unit 123")

    def test_unit_no_space_without_known_descriptors(self):
        self.addr.unit = "1A"
        self.assertEqual(self.addr.formatted_unit(), "Unit 1A")

    def test_unit_with_space_and_known_descriptor(self):
        self.addr.unit = "Apt 1A"
        self.assertEqual(self.addr.formatted_unit(), "Apt 1A")

    def test_bare_unit(self):
        self.addr.unit = "1A"
        self.assertEqual(self.addr.formatted_unit(bare=True), "1A")

    def test_floor_formatting(self):
        self.addr.floor = "5"
        self.assertEqual(self.addr.formatted_unit(), "Floor 5")

    def test_room_formatting(self):
        self.addr.room = "101"
        self.assertEqual(self.addr.formatted_unit(), "Room 101")


class TestALIndividual(unittest.TestCase):
    def setUp(self):
        self.individual = ALIndividual()

        # Creating a mock object for this_thread
        self.this_thread = Mock()

        # Mocking this_thread.global_vars.user as a distinct object
        self.this_thread.global_vars.user = Mock()

        # Assigning this_thread to self.individual
        self.individual.this_thread = self.this_thread

    def test_phone_numbers(self):
        self.assertEqual(self.individual.phone_numbers(), "")
        self.individual.phone_number = ""
        self.individual.mobile_number = None
        self.assertEqual(self.individual.phone_numbers(), "")

        self.individual.mobile_number = "aoueb"
        self.assertEqual(self.individual.phone_numbers(), "aoueb")
        self.individual.phone_number = "409_4567890 ext. 123"
        self.assertEqual(
            self.individual.phone_numbers(),
            "aoueb (cell), 409_4567890 ext. 123 (other)",
        )

        self.individual.mobile_number = "4094567890"
        self.individual.phone_number = None
        self.assertEqual(self.individual.phone_numbers(), "(409) 456-7890")
        self.individual.phone_number = "987-6543210"
        self.assertEqual(
            self.individual.phone_numbers(),
            "(409) 456-7890 (cell), (987) 654-3210 (other)",
        )

        # TODO: Not sure how to test international numbers
        # TODO: test impounding

    def test_pronoun(self):
        self.individual.pronouns = None
        self.individual.person_type = "individual"

        self.individual.gender = "female"
        self.assertEqual(self.individual.pronoun(), "her")
        self.assertEqual(self.individual.pronoun(capitalize=True), "Her")

        self.individual.gender = "male"
        self.assertEqual(self.individual.pronoun(), "him")
        self.assertEqual(self.individual.pronoun(capitalize=True), "Him")

        self.individual.gender = "nonbinary"
        self.assertEqual(self.individual.pronoun(), "them")
        self.assertEqual(self.individual.pronoun(capitalize=True), "Them")

        # Checks for preferred pronouns
        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": True,
                "he/him/his": False,
                "they/them/theirs": False,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun(), "her")
        self.assertEqual(self.individual.pronoun(capitalize=True), "Her")

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": True,
                "they/them/theirs": False,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun(), "him")
        self.assertEqual(self.individual.pronoun(capitalize=True), "Him")

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": False,
                "they/them/theirs": True,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun(), "them")
        self.assertEqual(self.individual.pronoun(capitalize=True), "Them")

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": False,
                "they/them/theirs": False,
                "ze/zir/zirs": True,
            },
        )
        self.assertEqual(self.individual.pronoun(), "zir")
        self.assertEqual(self.individual.pronoun(capitalize=True), "Zir")

        self.individual.pronouns = None
        self.individual.gender = None
        self.individual.person_type = "business"
        self.assertEqual(self.individual.pronoun(), "it")
        self.assertEqual(self.individual.pronoun(capitalize=True), "It")

    def test_pronoun_objective(self):
        self.individual.pronouns = None
        self.individual.person_type = "individual"

        self.individual.gender = "female"
        self.assertEqual(self.individual.pronoun_objective(), "her")
        self.assertEqual(self.individual.pronoun_objective(capitalize=True), "Her")

        self.individual.gender = "male"
        self.assertEqual(self.individual.pronoun_objective(), "him")
        self.assertEqual(self.individual.pronoun_objective(capitalize=True), "Him")

        self.individual.gender = "nonbinary"
        self.assertEqual(self.individual.pronoun_objective(), "them")
        self.assertEqual(self.individual.pronoun_objective(capitalize=True), "Them")

        # Checks for preferred pronouns
        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": True,
                "he/him/his": False,
                "they/them/theirs": False,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun_objective(), "her")
        self.assertEqual(self.individual.pronoun_objective(capitalize=True), "Her")

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": True,
                "they/them/theirs": False,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun_objective(), "him")
        self.assertEqual(self.individual.pronoun_objective(capitalize=True), "Him")

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": False,
                "they/them/theirs": True,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun_objective(), "them")
        self.assertEqual(self.individual.pronoun_objective(capitalize=True), "Them")

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": False,
                "they/them/theirs": False,
                "ze/zir/zirs": True,
            },
        )
        self.assertEqual(self.individual.pronoun_objective(), "zir")
        self.assertEqual(self.individual.pronoun_objective(capitalize=True), "Zir")

        self.individual.pronouns = None
        self.individual.gender = None
        self.individual.person_type = "business"
        self.assertEqual(self.individual.pronoun_objective(), "it")
        self.assertEqual(self.individual.pronoun_objective(capitalize=True), "It")

    def test_pronoun_possessive(self):
        self.individual.pronouns = None
        self.individual.person_type = "individual"
        item = "fish"

        self.individual.gender = "female"
        self.assertEqual(self.individual.pronoun_possessive(item), "her fish")
        self.assertEqual(
            self.individual.pronoun_possessive(item, capitalize=True), "Her fish"
        )

        self.individual.gender = "male"
        self.assertEqual(self.individual.pronoun_possessive(item), "his fish")
        self.assertEqual(
            self.individual.pronoun_possessive(item, capitalize=True), "His fish"
        )

        self.individual.gender = "nonbinary"
        self.assertEqual(self.individual.pronoun_possessive(item), "their fish")
        self.assertEqual(
            self.individual.pronoun_possessive(item, capitalize=True), "Their fish"
        )

        # Checks for preferred pronouns
        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": True,
                "he/him/his": False,
                "they/them/theirs": False,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun_possessive(item), "her fish")
        self.assertEqual(
            self.individual.pronoun_possessive(item, capitalize=True), "Her fish"
        )

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": True,
                "they/them/theirs": False,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun_possessive(item), "his fish")
        self.assertEqual(
            self.individual.pronoun_possessive(item, capitalize=True), "His fish"
        )

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": False,
                "they/them/theirs": True,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun_possessive(item), "their fish")
        self.assertEqual(
            self.individual.pronoun_possessive(item, capitalize=True), "Their fish"
        )

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": False,
                "they/them/theirs": False,
                "ze/zir/zirs": True,
            },
        )
        self.assertEqual(self.individual.pronoun_possessive(item), "zir fish")
        self.assertEqual(
            self.individual.pronoun_possessive(item, capitalize=True), "Zir fish"
        )

        self.individual.pronouns = None
        self.individual.gender = None
        self.individual.person_type = "business"
        self.assertEqual(self.individual.pronoun_possessive(item), "its fish")
        self.assertEqual(
            self.individual.pronoun_possessive(item, capitalize=True), "Its fish"
        )

    def test_pronoun_subjective(self):
        self.individual.pronouns = None
        self.individual.person_type = "individual"

        self.individual.gender = "female"
        self.assertEqual(self.individual.pronoun_subjective(), "she")
        self.assertEqual(self.individual.pronoun_subjective(capitalize=True), "She")

        self.individual.gender = "male"
        self.assertEqual(self.individual.pronoun_subjective(), "he")
        self.assertEqual(self.individual.pronoun_subjective(capitalize=True), "He")

        self.individual.gender = "nonbinary"
        self.assertEqual(self.individual.pronoun_subjective(), "they")
        self.assertEqual(self.individual.pronoun_subjective(capitalize=True), "They")

        # Checks for preferred pronouns
        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": True,
                "he/him/his": False,
                "they/them/theirs": False,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun_subjective(), "she")
        self.assertEqual(self.individual.pronoun_subjective(capitalize=True), "She")

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": True,
                "they/them/theirs": False,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun_subjective(), "he")
        self.assertEqual(self.individual.pronoun_subjective(capitalize=True), "He")

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": False,
                "they/them/theirs": True,
                "ze/zir/zirs": False,
            },
        )
        self.assertEqual(self.individual.pronoun_subjective(), "they")
        self.assertEqual(self.individual.pronoun_subjective(capitalize=True), "They")

        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": False,
                "they/them/theirs": False,
                "ze/zir/zirs": True,
            },
        )
        self.assertEqual(self.individual.pronoun_subjective(), "ze")
        self.assertEqual(self.individual.pronoun_subjective(capitalize=True), "Ze")

        self.individual.pronouns = None
        self.individual.gender = None
        self.individual.person_type = "business"
        self.assertEqual(self.individual.pronoun_subjective(), "it")
        self.assertEqual(self.individual.pronoun_subjective(capitalize=True), "It")

    def test_custom_pronouns(self):
        self.individual.pronouns = DADict(
            auto_gather=False,
            gathered=True,
            elements={
                "she/her/hers": False,
                "he/him/his": False,
                "they/them/theirs": False,
                "ze/zir/zirs": False,
                "self-described": True,
            },
        )

        self.individual.pronouns_self_described = "Xe/Xir/Xem"
        self.assertEqual(self.individual.pronoun_objective(), "xe")
        self.assertEqual(self.individual.pronoun_subjective(), "xir")
        self.assertEqual(self.individual.pronoun_possessive("fish"), "xem fish")

        self.individual.pronouns_self_described = "Xe/Xir/Xirs/xem/xirself"
        # Should raise an exception
        with self.assertRaises(DAAttributeError):
            self.individual.pronoun_objective()

    def test_name_methods(self):
        self.individual.name.first = "John"
        self.individual.name.middle = "Jacob"
        self.individual.name.last = "Jingleheimer"

        self.assertEqual(self.individual.name_full(), "John Jacob Jingleheimer")
        self.assertEqual(self.individual.name_initials(), "John J. Jingleheimer")
        self.assertEqual(self.individual.name_short(), "John Jingleheimer")


class test_get_visible_al_nav_items(unittest.TestCase):
    def test_case_1(self):
        data = [
            "Selecting_Legal_Category",
            {"personal_injury": "Gather Evidence", "hidden": False},
            {"divorce": "Collect Financial Documents", "hidden": False},
            {
                "business_dispute": [
                    "Interview Witnesses",
                    {"subtask": "Document Communication", "hidden": False},
                ]
            },
        ]
        expected = [
            "Selecting_Legal_Category",
            {"personal_injury": "Gather Evidence"},
            {"divorce": "Collect Financial Documents"},
            {
                "business_dispute": [
                    "Interview Witnesses",
                    {"subtask": "Document Communication"},
                ]
            },
        ]
        self.assertEqual(get_visible_al_nav_items(data), expected)

    def test_case_2(self):
        data = [
            "Drafting_Documents",
            {"pre_nup": "List Assets"},
            "Finding_Legal_Representative",
            {
                "business_formation": [
                    "Choose Business Type",
                    {"subtask": "Decide Ownership Structure", "hidden": True},
                ]
            },
            {"real_estate": "Draft Lease Agreement", "hidden": True},
        ]
        expected = [
            "Drafting_Documents",
            {"pre_nup": "List Assets"},
            "Finding_Legal_Representative",
            {"business_formation": ["Choose Business Type"]},
        ]
        self.assertEqual(get_visible_al_nav_items(data), expected)

    def test_case_3(self):
        data = [
            "Negotiating_Terms",
            "Reviewing_Contracts",
            {
                "estate_planning": [
                    "Draft Will",
                    {"subtask": "Designate Beneficiaries", "hidden": False},
                    {"subtask": "Determine Power of Attorney", "hidden": False},
                ]
            },
        ]
        expected = [
            "Negotiating_Terms",
            "Reviewing_Contracts",
            {
                "estate_planning": [
                    "Draft Will",
                    {"subtask": "Designate Beneficiaries"},
                    {"subtask": "Determine Power of Attorney"},
                ]
            },
        ]
        self.assertEqual(get_visible_al_nav_items(data), expected)

    def test_case_4(self):
        data = [
            {"employment_law": "Understand Employee Rights", "hidden": False},
            "Preparing_for_Court",
            {
                "criminal_defense": [
                    "Know Your Rights",
                    {"subtask": "Hire Defense Attorney", "hidden": True},
                ]
            },
        ]
        expected = [
            {"employment_law": "Understand Employee Rights"},
            "Preparing_for_Court",
            {"criminal_defense": ["Know Your Rights"]},
        ]
        self.assertEqual(get_visible_al_nav_items(data), expected)


if __name__ == "__main__":
    unittest.main()



from collections.abc import Iterable
from typing import List, Dict, Any, Optional, Set, Union, Optional
from docassemble.base.util import (
    all_variables,
    as_datetime,
    create_session,
    DADateTime,
    DAFile,
    DAFileCollection,
    DAFileList,
    DALazyTemplate,
    DAList,
    DAObject,
    DASet,
    format_time,
    get_config,
    get_language,
    get_default_timezone,
    get_session_variables,
    interview_menu,
    interview_url,
    log,
    set_parts,
    set_session_variables,
    set_variables,
    url_action,
    url_ask,
    user_has_privilege,
    user_info,
    user_logged_in,
    validation_error,
    variables_snapshot_connection,
    word,
)
from docassemble.webapp.db_object import init_sqlalchemy
from sqlalchemy.sql import text
from docassemble.base.functions import server, safe_json, serializable_dict
from .al_document import (
    ALDocument,
    ALDocumentBundle,
    ALExhibit,
    ALExhibitList,
    ALStaticDocument,
)
import json
import os
import re

try:
    import zoneinfo  # type: ignore
except ImportError:
    import backports.zoneinfo as zoneinfo  # type: ignore

__all__ = [
    "al_session_store_default_filename",
    "config_with_language_fallback",
    "delete_interview_sessions",
    "export_interview_variables",
    "get_filtered_session_variables_string",
    "get_filtered_session_variables",
    "get_interview_metadata",
    "get_saved_interview_list",
    "interview_list_html",
    "is_file_like",
    "is_valid_json",
    "load_interview_answers",
    "load_interview_json",
    "rename_current_session",
    "rename_interview_answers",
    "save_interview_answers",
    "session_list_html",
    "set_current_session_metadata",
    "set_interview_metadata",
]

db = init_sqlalchemy()

al_sessions_variables_to_remove: Set = {
    # Internal fields
    "_internal",
    "allow_cron",
    "device_local",
    "feedback_form",
    "github_repo_name",
    "github_user",
    "interview_short_title",
    "metadata_title",
    "multi_user",
    "nav",
    "session_local",
    "speak_text",
    "url_args",
    "user_local",
    # Database-like fields we don't need to copy
    "all_courts",
    "court_emails",
    "macourts",
    # AssemblyLine form-specific fields
    "al_form_type",
    "al_version",
    "form_approved_for_email_filing",
    "interview_metadata",
    "package_name",
    "package_version_number",
    "user_has_saved_answers",
    # Variables that should be calculated fresh
    "about_this_interview_version_info",
    "al_court_bundle",
    "AL_DEFAULT_COUNTRY",
    "AL_DEFAULT_LANGUAGE",
    "AL_DEFAULT_OVERFLOW_MESSAGE",
    "AL_DEFAULT_STATE",
    "al_enable_incomplete_downloads",
    "al_interview_languages",
    "al_logo",
    "AL_ORGANIZATION_HOMEPAGE",
    "AL_ORGANIZATION_TITLE",
    "al_terms_of_use",
    "al_user_bundle",
    "al_user_default_language",
    "al_user_language",
    "case_name",
    "enable_al_language",
    "signature_date",
    # Variables from saving/loading state
    "al_formatted_sessions",
    "al_session_store_default_filename",
    "al_sessions_copy_success",
    "al_sessions_fast_forward_filtered_vars",
    "al_sessions_fast_forward_session",
    "al_sessions_filtered_vars",
    "al_sessions_interview_title",
    "al_sessions_launch_new_session",
    "al_sessions_list",
    "al_sessions_new_session_id",
    "al_sessions_preview_variables",
    "al_sessions_save_session_snapshot_success",
    "al_sessions_save_session_snapshot",
    "al_sessions_snapshot_label",
    "al_sessions_snapshot_results",
    "al_sessions_source_session",
    "al_sessions_url_ask_fast_forward",
    "al_sessions_url_ask_snapshot",
    "al_sessions_variables_to_remove_from_new_interview",
    "al_sessions_variables_to_remove",
    "al_sessions_additional_variables_to_filter",
    "al_simple_filtered_vars",
    "filtered_vars_tmp",
    "is_file_like",
    "simple_filtered_vars_tmp",
    # Some type annotations from Typing that seem plausible we'll use (not everything)
    "Annotated",
    "Any",
    "AnyStr",
    "BinaryIO",
    "Callable",
    "ChainMap",
    "ClassVar",
    "Concatenate",
    "Counter",
    "DefaultDict",
    "Deque",
    "Dict",
    "Final",
    "FrozenSet",
    "Generic",
    "IO",
    "Iterable",
    "List",
    "Match",
    "NamedTuple",
    "NewType",
    "Optional",
    "OrderedDict",
    "ParamSpec",
    "Pattern",
    "Protocol",
    "Set",
    "Text",
    "TextIO",
    "Tuple",
    "TypedDict",
    "TypeGuard",
    "TypeLiteral",
    "TypeVar",
    "Union",
    # Variables that should always be created by code, so safe to recalculate
    "al_menu_items",
    "al_menu_items_default_items",
    "al_menu_items_custom_items",
    "menu_items",
    "user_role",
    "user_started_case",
}

al_sessions_variables_to_remove_from_new_interview = [
    "docket_number",
    "docket_numbers",
    "user_ask_role",
]

system_interviews: List[Dict[str, Any]] = interview_menu()


def _package_name(package_name: Optional[str] = None):
    """Get package name without the name of the current module, like: docassemble.ALWeaver instead of
    docassemble.ALWeaver.advertise_capabilities

    Args:
        package_name (str, optional): The package name to process. If `None`, will use the existing package name `__name__` instead

    Returns:
        str: The package name without the current module name.
    """
    if not package_name:
        package_name = __name__
    try:
        return ".".join(package_name.split(".")[:-1])
    except:
        return package_name


al_session_store_default_filename = f"{_package_name()}:al_saved_sessions_store.yml"


def is_file_like(obj: Any) -> bool:
    """
    Return True if the object is a file-like object.

    Args:
        obj (Any): The object to test

    Returns:
        bool: True if the object is a file-like object.
    """
    if isinstance(
        obj,
        (
            DAFile,
            DAFileCollection,
            DAFileList,
            ALDocument,
            ALDocumentBundle,
            ALStaticDocument,
            ALExhibit,
            ALExhibitList,
            DALazyTemplate,
        ),
    ):
        return True

    return False


def set_interview_metadata(
    filename: str, session_id: int, data: Dict, metadata_key_name="metadata"
) -> None:
    """Add searchable interview metadata for the specified filename and session ID.
    Intended to be used to add an interview title, etc.
    Standardized metadata dictionary:
    - title
    - subtitle
    - original_interview_filename
    - variable_count

    Args:
        filename (str): The filename of the interview to add metadata for
        session_id (int): The session ID of the interview to add metadata for
        data (Dict): The metadata to add
        metadata_key_name (str, optional): The name of the metadata key. Defaults to "metadata".
    """
    server.write_answer_json(
        session_id, filename, safe_json(data), tags=metadata_key_name, persistent=True
    )


def get_interview_metadata(
    filename: str, session_id: int, metadata_key_name: str = "metadata"
) -> Dict[str, Any]:
    """Retrieve the unencrypted metadata associated with an interview.
    We implement this with the docassemble jsonstorage table and a dedicated `tag` which defaults to `metadata`.

    Args:
        filename (str): The filename of the interview to retrieve metadata for
        session_id (int): The session ID of the interview to retrieve metadata for
        metadata_key_name (str, optional): The name of the metadata key. Defaults to "metadata".

    Returns:
        Dict[str, Any]: The metadata associated with the interview
    """
    conn = variables_snapshot_connection()
    with conn.cursor() as cur:
        query = "select data from jsonstorage where filename=%(filename)s and tags=%(tags)s and key=%(session_id)s"
        cur.execute(
            query,
            {"filename": filename, "tags": metadata_key_name, "session_id": session_id},
        )
        val = cur.fetchone()
    conn.close()
    if val and len(val):
        return val[0]  # cur.fetchone() returns a tuple
    return val or {}


def get_saved_interview_list(
    filename: Optional[str] = al_session_store_default_filename,
    user_id: Union[int, str, None] = None,
    metadata_key_name: str = "metadata",
    limit: int = 50,
    offset: int = 0,
    filename_to_exclude: str = "",
    exclude_current_filename: bool = True,
    exclude_filenames: Optional[List[str]] = None,
    exclude_newly_started_sessions: bool = False,
) -> List[Dict[str, Any]]:
    """Get a list of saved sessions for the specified filename. If the save_interview_answers function was used
    to add metadata, the result list will include columns containing the metadata.
    If the user is a developer or administrator, setting user_id = None will list all interviews on the server. Otherwise,
    the user is limited to their own sessions.

    Setting `exclude_newly_started_sessions` to True will exclude any results from the list that are still on
    "step 1". Note that while this may be useful to filter out interviews that were accidentally started
    and likely do not need to be resumed, it will also have the side effect of excluding all answer sets from the
    results. Answer sets generally have exactly one "step", which is the step where information was copied from
    an existing interview to the answer set.

    Args:
        filename (str, optional): The filename of the interview to retrieve sessions for. Defaults to al_session_store_default_filename.
        user_id (Union[int, str, None], optional): The user ID to retrieve sessions for. Defaults to None.
        metadata_key_name (str, optional): The name of the metadata key. Defaults to "metadata".
        limit (int, optional): The maximum number of results to return. Defaults to 50.
        offset (int, optional): The offset to start returning results from. Defaults to 0.
        filename_to_exclude (str, optional): The filename to exclude from the results. Defaults to "".
        exclude_current_filename (bool, optional): Whether to exclude the current filename from the results. Defaults to True.
        exclude_filenames (Optional[List[str]], optional): A list of filenames to exclude from the results. Defaults to None.
        exclude_newly_started_sessions (bool, optional): Whether to exclude sessions that are still on "step 1". Defaults to False.

    Returns:
        List[Dict[str, Any]]: A list of saved sessions for the specified filename.
    """
    # We use an `offset` instead of a cursor because it is simpler and clearer
    # while it appears to be performant enough for real-world usage.
    # Up to ~ 1,000 sessions performs well and is higher than expected for an end-user
    get_sessions_query = text(
        """
           SELECT  userdict.indexno
           ,userdict.filename as filename
           ,num_keys
           ,userdictkeys.user_id as user_id
           ,userdict.modtime as modtime
           ,userdict.key as key
           ,jsonstorage.data->'auto_title' as auto_title
           ,jsonstorage.data->'title' as title
           ,jsonstorage.data->'description' as description
           ,jsonstorage.data->'steps' as steps
           ,jsonstorage.data->'progress' as progress
           ,jsonstorage.data->'original_interview_filename' as original_interview_filename
           ,jsonstorage.data->'answer_count' as answer_count
           ,jsonstorage.data as data
    FROM userdict 
    NATURAL JOIN 
    (
      SELECT  key
             ,MAX(modtime) AS modtime
             ,COUNT(key)   AS num_keys
      FROM userdict
      GROUP BY  key
    ) mostrecent
    LEFT JOIN userdictkeys
    ON userdictkeys.key = userdict.key
    LEFT JOIN jsonstorage
    ON userdict.key = jsonstorage.key AND (jsonstorage.tags = :metadata)
    WHERE (userdictkeys.user_id = :user_id or :user_id is null)
    
    AND
    (userdict.filename = :filename OR :filename is null)
    AND (userdict.filename NOT IN :filenames_to_exclude)
    AND (NOT :exclude_newly_started_sessions OR num_keys > 1)
    ORDER BY modtime desc 
    LIMIT :limit
    OFFSET :offset;
    """
    )

    if offset < 0:
        offset = 0

    if not filename:
        filename = None  # Explicitly treat empty string as equivalent to None
    if exclude_current_filename:
        current_filename = user_info().filename
    else:
        current_filename = ""
    if not filename_to_exclude:
        filename_to_exclude = ""
    filenames_to_exclude = []
    if exclude_filenames:
        filenames_to_exclude.extend(exclude_filenames)
    filenames_to_exclude.extend([current_filename, filename_to_exclude])
    if user_id is None:
        if user_logged_in():
            user_id = user_info().id
        else:
            log("Asked to get interview list for user that is not logged in")
            return []

    if user_id == "all":
        if user_has_privilege(["developer", "admin"]):
            user_id = None
        elif user_logged_in():
            user_id = user_info().id
            log(
                f"User {user_info().email} does not have permission to list interview sessions belonging to other users"
            )
        else:
            log("Asked to get interview list for user that is not logged in")
            return []

    with db.connect() as con:
        rs = con.execute(
            get_sessions_query,
            {
                "metadata": metadata_key_name,
                "user_id": user_id,
                "filename": filename,
                "limit": limit,
                "offset": offset,
                "filenames_to_exclude": tuple(filenames_to_exclude),
                "exclude_newly_started_sessions": exclude_newly_started_sessions,
            },
        )
    sessions = []
    for session in rs:
        sessions.append(dict(session._mapping))

    return sessions


def delete_interview_sessions(
    user_id: Optional[int] = None,
    filename_to_exclude: str = al_session_store_default_filename,
    exclude_current_filename: bool = True,
) -> None:
    """
    Delete all sessions for the specified user, excluding the current filename
    and by default, the intentionally saved "answer sets". Created because
    interview_list(action="delete_all") is both quite slow and because it deletes answer sets.

    Args:
        user_id (Optional[int], optional): The user ID to delete sessions for. Defaults to None.
        filename_to_exclude (str, optional): The filename to exclude from the results. Defaults to al_session_store_default_filename.
        exclude_current_filename (bool, optional): Whether to exclude the current filename from the results. Defaults to True.
    """
    if not user_logged_in():
        log(
            "User that is not logged in does not have permission to delete any sessions"
        )
        return None
    delete_sessions_query = text(
        """
        DELETE FROM userdictkeys
        WHERE user_id = :user_id
        AND
        filename != :filename_to_exclude
        AND
        filename != :current_filename
        ;
        """
    )
    if not user_id:
        user_id = user_info().id
    if user_id != user_info().id and not user_has_privilege(["developer", "admin"]):
        user_id = user_info().id
        log(
            f"User {user_info().id} does not have permission to delete sessions that do not belong to them."
        )
    if exclude_current_filename:
        current_filename = user_info().filename or ""
    else:
        current_filename = ""
    if not filename_to_exclude:
        filename_to_exclude = ""

    log(f"Deleting sessions with {user_id} {filename_to_exclude} {current_filename}")

    with db.connect() as connection:
        connection.execute(
            delete_sessions_query,
            {
                "user_id": user_id,
                "filename_to_exclude": filename_to_exclude,
                "current_filename": current_filename,
            },
        )


def interview_list_html(
    filename: str = al_session_store_default_filename,
    user_id: Union[int, str, None] = None,
    metadata_key_name: str = "metadata",
    exclude_newly_started_sessions=False,
    # name_label: str = word("Title"),
    date_label: str = word("Date"),
    details_label: str = word("Details"),
    actions_label: str = word("Actions"),
    delete_label: str = word("Delete"),
    view_label: str = word("View"),
    load_action: str = "al_sessions_fast_forward_session",
    delete_action: str = "al_sessions_delete_session",
    view_only: bool = False,
    limit: int = 50,
    offset: int = 0,
    display_interview_title: bool = True,
    show_view_button: bool = True,
) -> str:
    """Return a string containing an HTML-formatted table with the list of saved answers
    associated with the specified filename.

    Designed to return a list of "answer sets" and by default clicking a title will
    trigger an action to load the answers into the current session. This only works as
    designed when inside an AssemblyLine line interview.

    `exclude_newly_started_sessions` should almost always be set to False, because most answer sets
    are on "page 1" (exactly 1 step was taken to copy the answers and the user isn't able to interact with the answer set
    itself in a way that adds additional steps)

    Args:
        filename (str, optional): Name of the file. Defaults to `al_session_store_default_filename`.
        user_id (Union[int, str, None], optional): User's ID. Defaults to None.
        metadata_key_name (str, optional): Name of the metadata key. Defaults to "metadata".
        exclude_newly_started_sessions (bool, optional): If True, newly started sessions are excluded. Defaults to False.
        date_label (str, optional): Label for the date column. Defaults to translated word "Date".
        details_label (str, optional): Label for the details column. Defaults to translated word "Details".
        actions_label (str, optional): Label for the actions column. Defaults to translated word "Actions".
        delete_label (str, optional): Label for the delete action. Defaults to translated word "Delete".
        view_label (str, optional): Label for the view action. Defaults to translated word "View".
        load_action (str, optional): Name of the load action. Defaults to "al_sessions_fast_forward_session".
        delete_action (str, optional): Name of the delete action. Defaults to "al_sessions_delete_session".
        view_only (bool, optional): If True, only view is allowed. Defaults to False.
        limit (int, optional): Limit for the number of sessions returned. Defaults to 50.
        offset (int, optional): Offset for the session list. Defaults to 0.
        display_interview_title (bool, optional): If True, displays the title of the interview. Defaults to True.
        show_view_button (bool, optional): If True, shows the view button. Defaults to True.

    Returns:
        str: HTML-formatted table containing the list of saved answers.
    """
    # TODO: Currently, using the `word()` function for translation, but templates
    # might be more flexible
    answers = get_saved_interview_list(
        filename=filename,
        user_id=user_id,
        metadata_key_name=metadata_key_name,
        limit=limit,
        offset=offset,
        exclude_current_filename=False,
        exclude_newly_started_sessions=exclude_newly_started_sessions,
    )

    if not answers:
        return ""

    table = '<div class="table-responsive"><table class="table table-striped al-saved-answer-table text-break">'
    table += f"""
    <thead>
      <th scope="col">
        &nbsp;
      </th>
      <th scope="col">{ date_label }</th>
      <th scope="col">{ details_label }</th>
      <th scope="col">{ actions_label }</th>
      </th>
    </thead>
    <tbody>
"""

    for answer in answers:
        answer = dict(answer)
        # Never display the current interview session
        if answer.get("key") == user_info().session:
            continue
        table += """<tr class="al-saved-answer-table-row">"""
        if view_only:
            table += f"""
            <td>{ nice_interview_subtitle(answer) or nice_interview_title(answer) }</td>
            """
        else:
            table += f"""
            <td class="text-break"><a href="{ url_action(load_action, i=answer.get("filename"), session=answer.get("key")) }">{ nice_interview_subtitle(answer) or nice_interview_title(answer) }</a></td>
            """
        table += f"""
        <td>{ as_datetime(answer.get("modtime")) }</td>
        <td>Page { answer.get("steps") or answer.get("num_keys") }"""
        if display_interview_title:
            table += f"""
                 <br/>{answer.get("original_interview_filename") or answer.get("filename") or "" }
            """
        table += f"""
        </td>
        <td>
          <a href="{ url_action(delete_action, filename=answer.get("filename"), session=answer.get("key")) }"><i class="far fa-trash-alt" title="{ delete_label }" aria-hidden="true"></i><span class="sr-only">{ delete_label }</span></a>
          """
        if show_view_button:
            table += f"""
                <a target="_blank" href="{ interview_url(i=answer.get("filename"), session=answer.get("key")) }">
                    <i class="far fa-eye" aria-hidden="true" title="{ view_label }"></i>
                    <span class="sr-only">{ view_label }</span>
                </a>
            """
        table += """
        </td>
        """
        table += "</tr>"
    table += "</tbody></table></div>"

    return table


def nice_interview_title(
    answer: Dict[str, str],
) -> str:
    """
    Return a human readable version of the interview name. Will try several strategies
    in descending priority order.
    1. Try looking up the interview title from the `dispatch` directive
    1. Try removing the package and path from the filename and replace _ with spaces.
    4. Finally, return "Untitled interview" or translated phrase from system-wide words.yml

    Args:
        answer (Dict[str, str]): The answer dictionary to get the interview title from

    Returns:
        str: The human readable interview title
    """
    if answer.get("filename"):
        for interview in system_interviews:
            if answer.get("filename") == interview.get("filename") and interview.get(
                "title"
            ):
                return interview["title"]
        filename = os.path.splitext(os.path.basename(answer.get("filename", "")))[0]
        if ":" in filename:
            filename = filename.split(":")[1]
        return filename.replace("_", " ").capitalize()
    else:
        return word("Untitled interview")


def pascal_to_zwspace(text: str) -> str:
    """
    Insert a zero-width space into words that are PascalCased to help
    with word breaks on small viewports.

    Args:
        text (str): The text to insert zero-width spaces into

    Returns:
        str: The text with zero-width spaces inserted
    """
    re_outer = re.compile(r"([^A-Z ])([A-Z])")
    re_inner = re.compile(r"(?<!^)([A-Z])([^A-Z])")
    return re_outer.sub(r"\1​\2", re_inner.sub(r"​\1\2", text))


def nice_interview_subtitle(answer: Dict[str, str], exclude_identical=True) -> str:
    """
    Return first defined of the "title" metadata, the "auto_title" metadata, or empty string.

    If exclude_identical, return empty string when title is the same as the subtitle.

    Args:
        answer (Dict[str, str]): The answer dictionary to get the interview subtitle from
        exclude_identical (bool, optional): If True, excludes the subtitle if it is identical to the title. Defaults to True.

    Returns:
        str: The human readable interview subtitle
    """
    if answer.get("title"):
        return pascal_to_zwspace(answer["title"])
    elif answer.get("auto_title") and (
        not exclude_identical
        or not (
            answer.get("auto_title", "").lower() == nice_interview_title(answer).lower()
        )
    ):
        return pascal_to_zwspace(answer["auto_title"])
    return ""


def radial_progress(answer: Dict[str, Union[str, int]]) -> str:
    """
    Return HTML for a radial progress bar, or the number of steps if progress isn't available in the metadata.

    Args:
        answer (Dict[str, Union[str, int]]): The answer dictionary to get the interview progress from

    Returns:
        str: the HTML as a string
    """
    if not answer.get("progress"):
        return f"Page {answer.get('steps') or answer.get('num_keys') or 1}"

    # For simulation purposes, assume a form is complete at page 30
    progress: int = (
        int(answer["progress"])
        or int(answer["steps"])
        or (int(answer["num_keys"]) or 1) * 4
    )

    # Get a number divisible by 10 and between 0 and 100 (percentage)
    nearest_10 = min(max(round(progress / 10) * 10, 0), 100)
    return f"""
      <div class="radialProgressBar progress-{ nearest_10 }">
          <div class="overlay">{ nearest_10 }%</div>
        </div>
    """


def local_date(utcstring: Optional[str]) -> DADateTime:
    """
    Return a localized date from a UTC string.

    Args:
        utcstring (Optional[str]): The UTC string to convert to a localized date

    Returns:
        DADateTime: The localized date
    """
    if not utcstring:
        return DADateTime()
    return (
        as_datetime(utcstring)
        .replace(tzinfo=zoneinfo.ZoneInfo("UTC"))
        .astimezone(zoneinfo.ZoneInfo(get_default_timezone()))
    )


def session_list_html(
    filename: Optional[str] = None,
    user_id: Union[int, str, None] = None,
    metadata_key_name: str = "metadata",
    filename_to_exclude: str = al_session_store_default_filename,
    exclude_current_filename: bool = True,
    exclude_filenames: Optional[List[str]] = None,
    exclude_newly_started_sessions: bool = False,
    name_label: str = word("Title"),
    date_label: str = word("Date modified"),
    details_label: str = word("Progress"),
    actions_label: str = word("Actions"),
    delete_label: str = word("Delete"),
    rename_label: str = word("Rename"),
    rename_action: str = "interview_list_rename_action",
    delete_action: str = "interview_list_delete_session",
    copy_action: str = "interview_list_copy_action",
    clone_label: str = word("Copy as answer set"),
    show_title: bool = True,
    show_copy_button: bool = True,
    limit: int = 50,
    offset: int = 0,
) -> str:
    """Return a string containing an HTML-formatted table with the list of user sessions.
    While interview_list_html() is for answer sets, this feature is for standard
    user sessions. The results exclude the answer set filename by default.

    Args:
        filename (Optional[str], optional): Name of the file. Defaults to None.
        user_id (Union[int, str, None], optional): User's ID. Defaults to None.
        metadata_key_name (str, optional): Name of the metadata key. Defaults to "metadata".
        filename_to_exclude (str, optional): Name of the file to exclude. Defaults to `al_session_store_default_filename`.
        exclude_current_filename (bool, optional): If True, excludes the current filename. Defaults to True.
        exclude_filenames (Optional[List[str]], optional): List of filenames to exclude. Defaults to None.
        exclude_newly_started_sessions (bool, optional): If True, excludes newly started sessions. Defaults to False.
        name_label (str, optional): Label for the session name/title. Defaults to translated word "Title".
        date_label (str, optional): Label for the date column. Defaults to translated word "Date modified".
        details_label (str, optional): Label describing the progress of the session. Defaults to translated word "Progress".
        actions_label (str, optional): Label for actions applicable to the session. Defaults to translated word "Actions".
        delete_label (str, optional): Label for the delete action. Defaults to translated word "Delete".
        rename_label (str, optional): Label for the rename action. Defaults to translated word "Rename".
        rename_action (str, optional): Name of the rename action. Defaults to "interview_list_rename_action".
        delete_action (str, optional): Name of the delete action. Defaults to "interview_list_delete_session".
        copy_action (str, optional): Name of the copy action. Defaults to "interview_list_copy_action".
        clone_label (str, optional): Label for the action to copy as an answer set. Defaults to translated word "Copy as answer set".
        show_title (bool, optional): If True, shows the title of the session. Defaults to True.
        show_copy_button (bool, optional): If True, show a copy button for answer sets. Defaults to True.
        limit (int, optional): Limit for the number of sessions returned. Defaults to 50.
        offset (int, optional): Offset for the session list. Defaults to 0.

    Returns:
        str: HTML-formatted table containing the list of user sessions.
    """

    # TODO: think through how to translate this function. Templates probably work best but aren't
    # convenient to pass around
    answers = get_saved_interview_list(
        filename=filename,
        user_id=user_id,
        metadata_key_name=metadata_key_name,
        limit=limit,
        offset=offset,
        filename_to_exclude=filename_to_exclude,
        exclude_current_filename=exclude_current_filename,
        exclude_filenames=exclude_filenames,
        exclude_newly_started_sessions=exclude_newly_started_sessions,
    )

    if not answers:
        return ""

    table = '<div class="table-responsive"><table class="table table-striped al-saved-answer-table">'
    table += f"""
    <thead>
      <th scope="col">
       { name_label }
      </th>
      <th scope="col">{ date_label }</th>
      <th scope="col">{ details_label }</th>
      <th scope="col">{ actions_label }</th>
      </th>
    </thead>
    <tbody>
"""

    for answer in answers:
        answer = dict(answer)
        # Never display the current interview session
        if answer.get("key") == user_info().session:
            continue
        url_ask_rename = url_ask(
            [
                {"undefine": ["al_sessions_snapshot_new_label"]},
                {
                    "action": rename_action,
                    "arguments": {
                        "session": answer.get("key"),
                        "filename": answer.get("filename"),
                        "title": nice_interview_subtitle(answer)
                        or nice_interview_title(answer),
                    },
                },
            ]
        )
        url_ask_copy = url_ask(
            [
                {"undefine": ["al_sessions_copy_as_answer_set_label"]},
                {
                    "action": copy_action,
                    "arguments": {
                        "session": answer.get("key"),
                        "filename": answer.get("filename"),
                        "title": answer.get("title") or "",
                        "original_interview_filename": nice_interview_title(answer),
                    },
                },
            ]
        )
        url_ask_delete = url_ask(
            [
                {
                    "action": delete_action,
                    "arguments": {
                        "session": answer.get("key"),
                        "filename": answer.get("filename"),
                    },
                }
            ]
        )

        table += """<tr class="al-saved-answer-table-row">"""
        if show_title:
            table += f"""
            <td class="text-break">
            <a class="al-session-form-title" href="{ interview_url(i=answer.get("filename"), session=answer.get("key")) }">{ nice_interview_title(answer) }</a>
            {"<br/>" if nice_interview_subtitle(answer) else ""}
            <span class="al-session-form-subtitle">{ nice_interview_subtitle(answer) if nice_interview_subtitle(answer) else "" }</span>
            </td>
            """
        else:
            table += f"""
            <td class="text-break">
            <a class="al-session-form-title" href="{ interview_url(i=answer.get("filename"), session=answer.get("key")) }">{ nice_interview_subtitle(answer) or nice_interview_title(answer) }</a>
            </td>
            """
        table += f"""
        <td>
            <span class="al-session-date-modified">{ local_date(answer.get("modtime")) }</span> <br/>
            <span class="al-session-time-modified">{ format_time(local_date(answer.get("modtime")).time(), format="h:mm a") }</span>
        </td>
        <td class="al-progress-box">{ radial_progress(answer) }
        </td>
        <td>
          <a class="al-sessions-action-rename" href="{ url_ask_rename }"><i class="fa-solid fa-tag" aria-hidden="true" title="{ rename_label }"></i><span class="sr-only">{ rename_label }</span></a>
        """
        if (
            get_config("assembly line", {}).get("enable answer sets")
            and show_copy_button
        ):
            table += f"""
          &nbsp;
          <a class="al-sessions-actions-clone" href="{ url_ask_copy }"><i class="fa-regular fa-clone" aria-hidden="true" title="{clone_label}"></i><span class="sr-only">{ clone_label }</span></a>
          """
        table += f"""
          &nbsp;
          <a href="{ url_ask_delete }" class="text-danger al-delete-session"><i class="far fa-trash-alt text-danger" title="{ delete_label }" aria-hidden="true"></i><span class="sr-only">{ delete_label }</span></a>
        </td>
        """
        table += "</tr>"
    table += "</tbody></table></div>"

    return table


def rename_interview_answers(
    filename: str,
    session_id: int,
    new_name: str,
    metadata_key_name: str = "metadata",
) -> None:
    """Update the 'title' metadata of an interview, as stored in the dedicated `metadata` column, without touching other
    metadata that may be present.

    Args:
        filename (str): The filename of the interview to rename
        session_id (int): The session ID of the interview to rename
        new_name (str): The new name to set for the interview
        metadata_key_name (str, optional): The name of the metadata key. Defaults to "metadata".

    If exception is raised in set_session_variables, this will silently fail but log the error.
    """
    existing_metadata = get_interview_metadata(
        filename, session_id, metadata_key_name=metadata_key_name
    )
    existing_metadata["title"] = new_name
    set_interview_metadata(
        filename, session_id, existing_metadata, metadata_key_name=metadata_key_name
    )
    if session_id == user_info().session:
        set_parts(subtitle=new_name)
    else:
        try:
            set_session_variables(
                filename,
                session_id,
                {"_internal['subtitle']": new_name},
                overwrite=True,
            )
        except:
            log(
                f"Unable to update internal interview subtitle for session {filename}:{session_id} with new name {new_name}"
            )


def set_current_session_metadata(
    data: Dict[str, Any],
    metadata_key_name: str = "metadata",
) -> None:
    """
    Set metadata for the current session, such as the title, in an unencrypted database entry.

    Args:
        data (Dict[str, Any]): The metadata to set
        metadata_key_name (str, optional): The name of the metadata key. Defaults to "metadata".
    """
    return set_interview_metadata(
        user_info().filename,
        user_info().session,
        data,
        metadata_key_name=metadata_key_name,
    )


def rename_current_session(
    new_name: str,
    metadata_key_name: str = "metadata",
) -> None:
    """Update the "title" metadata entry for the current session without changing any other
    metadata that might be present.

    Args:
        new_name (str): The new name to set for the interview
        metadata_key_name (str, optional): The name of the metadata key. Defaults to "metadata".
    """
    return rename_interview_answers(
        user_info().filename,
        user_info().session,
        new_name,
        metadata_key_name=metadata_key_name,
    )


def save_interview_answers(
    filename: str = al_session_store_default_filename,
    variables_to_filter: Union[Set[str], List[str], None] = None,
    metadata: Optional[Dict] = None,
    metadata_key_name: str = "metadata",
    original_interview_filename=None,
    source_filename=None,
    source_session=None,
    additional_variables_to_filter=None,
) -> str:
    """
    Copies the answers from a given session into a new session with a specified interview filename.

    Args:
        filename (str, optional): The desired filename for the new session. Defaults to `al_session_store_default_filename`.
        variables_to_filter (Union[Set[str], List[str], None], optional): The "base" list or set of variables to filter out. Defaults to `al_sessions_variables_to_remove`. There's usually no reason to change this and changing it might break sessions.
        metadata (Optional[Dict], optional): Dictionary containing metadata. Defaults to an empty dictionary.
        metadata_key_name (str, optional): Key name for metadata storage. Defaults to "metadata".
        original_interview_filename (str, optional): Original filename of the interview. Defaults to None.
        source_filename (str, optional): Source filename to get session variables from. Defaults to None.
        source_session (str, optional): Session ID of the source file. Defaults to None.
        additional_variables_to_filter (Union[Set[str], List[str], None], optional): List or set of variables to filter out. Defaults to None.

    Returns:
        str: ID of the new session.
    """
    # Avoid using mutable default parameter
    if not variables_to_filter:
        variables_to_filter = al_sessions_variables_to_remove
    if not additional_variables_to_filter:
        additional_variables_to_filter = []
    if not metadata:
        metadata = {}

    # Get session variables, from either the specified or current session
    all_vars = get_filtered_session_variables(
        filename=source_filename,
        session_id=source_session,
        variables_to_filter=variables_to_filter,
        additional_variables_to_filter=additional_variables_to_filter,
    )

    try:
        # Sometimes include_internal breaks things
        metadata["steps"] = (
            all_variables(include_internal=True).get("_internal").get("steps", -1)
        )
    except:
        metadata["steps"] = -1

    if original_interview_filename:
        metadata["original_interview_filename"] = original_interview_filename
    else:
        metadata["original_interview_filename"] = all_variables(special="metadata").get(
            "title", user_info().filename.replace(":", " ").replace(".", " ")
        )
    metadata["answer_count"] = len(all_vars)

    # Create a new session
    new_session_id = create_session(filename)

    # Copy in the variables from the specified session
    set_session_variables(filename, new_session_id, all_vars, overwrite=True)

    # Add the metadata
    set_interview_metadata(
        filename, new_session_id, metadata, metadata_key_name=metadata_key_name
    )
    # Make the title display as the subtitle on the "My interviews" page
    if metadata.get("title"):
        try:
            set_session_variables(
                filename,
                new_session_id,
                {"_internal['subtitle']": metadata.get("title")},
                overwrite=True,
            )
        except:
            log(
                f"Unable to set internal interview subtitle for session {filename}:{new_session_id} with name {metadata.get('title')}"
            )

    return new_session_id


def get_filtered_session_variables(
    filename: Optional[str] = None,
    session_id: Optional[int] = None,
    variables_to_filter: Optional[Union[Set[str], List[str]]] = None,
    additional_variables_to_filter: Optional[Union[Set[str], List[str]]] = None,
) -> Dict[str, Any]:
    """
    Retrieves a filtered subset of variables from a specified interview and session.
    If no filename and session ID are given, it will return a filtered list of variables
    from the current interview.

    Args:
        filename (Optional[str], optional): Filename of the session. Defaults to None.
        session_id (Optional[int], optional): Session ID to retrieve variables from. Defaults to None.
        variables_to_filter (Union[Set[str], List[str], None], optional): List or set of variables to exclude. Defaults to `al_sessions_variables_to_remove`.
        additional_variables_to_filter (Union[Set[str], List[str], None], optional): List or set of additional variables to exclude. Defaults to None.

    Returns:
        Dict[str, Any]: A dictionary of filtered session variables.
    """
    if not variables_to_filter:
        variables_to_filter = al_sessions_variables_to_remove
    if not additional_variables_to_filter:
        additional_variables_to_filter = []
    variables_to_filter = set(variables_to_filter).union(
        set(additional_variables_to_filter)
    )

    if filename and session_id:
        all_vars = get_session_variables(filename, session_id, simplify=False)
    else:
        all_vars = all_variables(simplify=False, make_copy=True)

    all_vars = {k: v for k, v in all_vars.items() if k not in variables_to_filter}

    items_to_check = list(all_vars.items())

    while items_to_check:
        key, value = items_to_check.pop()

        # This condition only will apply to "top level" variables
        if is_file_like(value):
            del all_vars[key]
            continue

        if isinstance(value, DAObject):
            # docassemble overrides both __dir__ and __getattr__ for reasons unknown
            # we need to use the base Python versions to get what we expect
            attr_list = list(
                value.__dict__.keys()
            )  # skip over properties etc. vs using object.dir()
            for attr in attr_list:
                attr_val = object.__getattribute__(value, attr)
                if is_file_like(attr_val):
                    delattr(value, attr)
                elif isinstance(attr_val, (DAList, DASet, DAObject)):
                    items_to_check.append(
                        (None, attr_val)
                    )  # mimic dict.items() but the key isn't used

        if isinstance(value, (DAList, DASet)):
            new_elements = []
            for subitem in value.elements:
                if not is_file_like(subitem):
                    new_elements.append(subitem)
                    if isinstance(subitem, (DAList, DASet, DAObject)):
                        items_to_check.append((None, subitem))
            value.elements = (
                new_elements if isinstance(value, DAList) else set(new_elements)
            )

    return all_vars


def get_filtered_session_variables_string(
    filename: Optional[str] = None,
    session_id: Optional[int] = None,
    variables_to_filter: Union[Set[str], List[str], None] = None,
    additional_variables_to_filter: Optional[Union[Set[str], List[str]]] = None,
) -> str:
    """
    Returns a JSON string that represents the filtered contents of a specified filename and session ID.
    If no filename and session ID are provided, the current session's variables will be used.

    Args:
        filename (Optional[str], optional): Filename of the session. Defaults to None.
        session_id (Optional[int], optional): Session ID to retrieve variables from. Defaults to None.
        variables_to_filter (Union[Set[str], List[str], None], optional): List or set of variables to exclude. Defaults to `al_sessions_variables_to_remove`.
        additional_variables_to_filter (Union[Set[str], List[str], None], optional): List or set of additional variables to exclude. Defaults to None.

    Returns:
        str: A JSON-formatted string of filtered session variables.
    """
    simple_vars = serializable_dict(
        get_filtered_session_variables(
            filename=filename,
            session_id=session_id,
            variables_to_filter=variables_to_filter,
            additional_variables_to_filter=additional_variables_to_filter,
        )
    )
    return json.dumps(simple_vars)


def load_interview_answers(
    old_interview_filename: str,
    old_session_id: int,
    new_session: bool = False,
    new_interview_filename: Optional[str] = None,
    variables_to_filter: Optional[List[str]] = None,
    additional_variables_to_filter: Optional[List[str]] = None,
) -> Optional[Union[int, bool]]:
    """
    Loads answers from a specified session. If the parameter `new_session` is set to True, it will create
    a new session with the provided or current interview filename. Otherwise, it will load the answers into
    the active session. This function is primarily used for migrating answers between sessions.

    Args:
        old_interview_filename (str): Filename of the old interview.
        old_session_id (int): Session ID of the old interview.
        new_session (bool, optional): Determines whether to create a new session. Defaults to False.
        new_interview_filename (Optional[str], optional): Filename for the new session. Defaults to None.
        variables_to_filter (Optional[List[str]], optional): List of variables to exclude. Defaults to None.
        additional_variables_to_filter (Optional[List[str]], optional): List of additional variables to exclude. Defaults to None.

    Returns:
        Optional[Union[int, bool]]: ID of the newly created session if `new_session` is True, otherwise True or False based on success.
    """

    old_variables = get_filtered_session_variables(
        filename=old_interview_filename,
        session_id=old_session_id,
        variables_to_filter=variables_to_filter,
        additional_variables_to_filter=additional_variables_to_filter,
    )

    if new_session:
        if not new_interview_filename:
            new_interview_filename = user_info().filename
        new_session_id = create_session(new_interview_filename)
        set_session_variables(new_interview_filename, new_session_id, old_variables)
        return new_session_id
    else:
        try:
            set_variables(old_variables)
            return True
        except:
            return False


def load_interview_json(
    json_string: str,
    new_session: bool = False,
    new_interview_filename: Optional[str] = None,
    variables_to_filter: Optional[List[str]] = None,
) -> Optional[int]:
    """
    Given a JSON string, this function loads the specified variables into a Docassemble session.
    JSON strings containing annotated class names will be transformed into Docassemble objects.
    If the `new_session` argument is not set, the JSON answers will be loaded into the current interview.

    Args:
        json_string (str): A JSON-formatted string containing session variables.
        new_session (bool, optional): Specifies whether to create a new session or load into the current one. Defaults to False.
        new_interview_filename (Optional[str], optional): Filename for the new session. Defaults to None.
        variables_to_filter (Optional[List[str]], optional): List of variables to exclude. Defaults to None.

    Returns:
        Optional[Union[int, bool]]: ID of the newly created session if `new_session` is True, otherwise True or False based on success.
    """
    json_processed = json.loads(json_string)

    if new_session:
        if not new_interview_filename:
            new_interview_filename = user_info().filename
        new_session_id = create_session(new_interview_filename)
        set_session_variables(
            new_interview_filename, new_session_id, json_processed, process_objects=True
        )
        return new_session_id
    else:
        try:
            set_variables(json_processed, process_objects=True)
            return True
        except:
            return False


def export_interview_variables(
    filename: Optional[str] = None,
    session_id: Optional[int] = None,
    variables_to_filter: Union[Set, List[str], None] = None,
    output: DAFile = None,
    additional_variables_to_filter: Union[Set, List[str], None] = None,
) -> DAFile:
    """
    Generates a DAFile containing a JSON representation of a specified session's interview answers.
    The resultant output is compatible with `set_session_variables(process_objects=True)` and
    `set_variables(process_objects=True)` methods.

    Args:
        filename (Optional[str], optional): Filename of the session. Defaults to None.
        session_id (Optional[int], optional): Session ID to retrieve variables from. Defaults to None.
        variables_to_filter (Union[Set, List[str], None], optional): List or set of variables to exclude. Defaults to None.
        output (DAFile, optional): DAFile to write the JSON output to. If None, a new DAFile is created.
        additional_variables_to_filter (Union[Set, List[str], None], optional): List or set of additional variables to exclude. Defaults to None.

    Returns:
        DAFile: DAFile with a JSON representation of the answers
    """
    if not output:
        output = DAFile()
    output.initialize(filename="variables.json")
    variables_string = get_filtered_session_variables_string(
        filename=filename,
        session_id=session_id,
        variables_to_filter=variables_to_filter,
        additional_variables_to_filter=additional_variables_to_filter,
    )

    output.write(variables_string)
    output.commit()
    return output


def is_valid_json(json_string: str) -> bool:
    """
    Checks if the provided string is a valid JSON-formatted string.

    Args:
        json_string (str): The string to be checked for JSON validity.

    Returns:
        bool: True if the string is a valid JSON, otherwise it raises a validation error and returns False.
    """
    try:
        json.loads(json_string)
    except:
        validation_error("Enter a valid JSON-formatted string")
        return False
    return True


def config_with_language_fallback(
    config_key: str, top_level_config_key: Optional[str] = None
) -> Optional[str]:
    """Returns the value of a config key under `assembly line` `interview list` with options to fallback
    to an alternative key at the top level of the global configuration.

    Used in interview_list.yml to allow overriding some of the labels in the interview list
    with options specified in the global configuration. top_level_config should be reserved
    to handle backwards compatibility (e.g., changed location of some configuration keys)

    Example configuration, showing both the single-string and language-specific string options:
        assembly line:
            interview list:
                title:
                    en: In progress forms
                    es: Formularios en progreso
                short title: My forms

    Args:
        config_key (str): The config key to look up. The config can be a single string or a dictionary with language keys.
        top_level_config_key (str, optional): Optional, alternative top-level config key to look up. Defaults to None.

    Returns:
        str: The value of the config key, or the alternative key, or None.
    """
    interview_list_config = get_config("assembly line", {}).get("interview list", {})
    if interview_list_config.get(config_key):
        if isinstance(interview_list_config.get(config_key), dict):
            if get_language() in interview_list_config.get(config_key):
                return interview_list_config.get(config_key)[get_language()]
            else:
                return next(iter(interview_list_config.get(config_key).values()), None)
        else:
            return interview_list_config.get(config_key)
    else:
        return get_config(top_level_config_key or config_key)
